
import os
import unittest
import numpy as np
from xrd_simulator.motion import RigidBodyMotion
import pkg_resources

class TestPhase(unittest.TestCase):

    def setUp(self):
        np.random.seed(1)
        
    def test_init(self):
        rotation_axis = np.random.rand(3,)
        rotation_axis = rotation_axis / np.linalg.norm(rotation_axis)
        rotation_angle = np.random.rand(1,)*np.pi
        translation = np.random.rand(3,)
        motion  = RigidBodyMotion(rotation_axis, rotation_angle, translation)

    def test_rotate(self):
        rotation_axis = np.array([1.,0,0])
        rotation_angle = np.pi/2.
        translation = np.random.rand(3,)
        motion = RigidBodyMotion(rotation_axis, rotation_angle, translation)
        
        x = motion.rotate( np.array([1.,0,0]), time=1)
        z = motion.rotate( np.array([0,1,0]), time=1)
        y = motion.rotate( np.array([0,0,-1]), time=1)

        self.assertAlmostEqual( np.linalg.norm(x-np.array([1.,0,0])), 0, msg='Error in rotator' )
        self.assertAlmostEqual( np.linalg.norm(y-np.array([0,1,0])), 0, msg='Error in rotator' )
        self.assertAlmostEqual( np.linalg.norm(z-np.array([0,0,1])), 0, msg='Error in rotator' )

    def test_translate(self):
        rotation_axis = np.random.rand(3,)
        rotation_axis = rotation_axis / np.linalg.norm(rotation_axis)
        rotation_angle = np.random.rand(1,)
        translation = np.random.rand(3,)-0.5
        motion = RigidBodyMotion(rotation_axis, rotation_angle, translation)

        v0 = np.random.rand(3,)-0.5
        v = motion.translate( v0, time=0.4323)

        self.assertAlmostEqual( np.linalg.norm(v-translation*0.4323-v0), 0, msg='Error in translation' )

    def test_call(self):
        rotation_axis = np.array([1.,0,0])
        rotation_angle = np.pi/2.
        translation = np.array([1.,0,0])
        motion = RigidBodyMotion(rotation_axis, rotation_angle, translation)
        
        y = np.array([0,1,0])
        v = motion( y, time=1)

        self.assertAlmostEqual( np.linalg.norm(v - np.array([1,0,1])), 0, msg='Error in rigid body transformation' )

if __name__ == '__main__':
    unittest.main()



"""Collection of functions for solving the Laue equations for package specific parametrization.
"""

import numpy as np

def get_G( U, B, G_hkl ):
    """Compute the diffraction vector G=UBG_HKL

    Args:
        U (:obj:`numpy array`) Orientation matrix of ``shape=(3,3)`` (unitary).
        B (:obj:`numpy array`): Reciprocal to grain coordinate mapping matrix of ``shape=(3,3)``.
        G_hkl (:obj:`numpy array`): Miller indices, i.e the h,k,l integers (``shape=(3,n)``).


    Returns:
        G (:obj:`numpy array`): Sample coordinate system diffraction vector. (``shape=(3,n)``)

    """
    return np.dot( np.dot(U, B), G_hkl  )

def get_bragg_angle( G, wavelength ):
    """Compute a Bragg angle given a diffraction (scattering) vector.

    Args:
        G (:obj:`numpy array`): Sample coordinate system diffraction vector. (``shape=(3,n)``)
        wavelength (:obj:`float`): Photon wavelength in units of angstrom.

    Returns:
        Bragg angles (:obj:`float`): in units of radians. (``shape=(n,)``)
        
    """
    return np.arcsin( np.linalg.norm(G,axis=0)*wavelength/(4*np.pi) )

def get_sin_theta_and_norm_G( G, wavelength ):
    """Compute a Bragg angle given a diffraction (scattering) vector.

    Args:
        G (:obj:`numpy array`): Sample coordinate system diffraction vector.
        wavelength (:obj:`float`): Photon wavelength in units of angstrom.

    Returns:
        sin(Bragg angle) (:obj:`float`): in units of radians and ||G||.
        
    """
    normG = np.linalg.norm(G,axis=0)
    return normG*wavelength/(4*np.pi), normG

def get_tangens_half_angle_equation(k1, theta, G, rhat ):
    """Find coefficient to the equation

    .. math::
        c_0 \cos(s \\alpha) + c_1 \sin(s \\alpha) + c_2 = 0. \\quad\\quad (1)

    """
    c_0  = np.dot( k1, G )
    c_1  = np.dot( np.cross( rhat , k1 ), G )
    c_2  = np.linalg.norm(k1) * np.linalg.norm(G) * np.sin(theta)
    return c_0, c_1, c_2

def find_solutions_to_tangens_half_angle_equation( c_0, c_1, c_2, alpha ):
    """Find all solutions, :obj:`s`, to the equation (maximum 2 solutions exists)

    .. math::
        c_0 \cos(s \\alpha) + c_1 \sin(s \\alpha) + c_2 = 0. \\quad\\quad (1)

    by rewriting as

    .. math::
        (c_2 - c_0) t^2 + 2 c_1 t + (c_0 + c_2) = 0. \\quad\\quad (2)

    where

    .. math::
        t = \\tan(s \\alpha / 2). \\quad\\quad (3)
    
    and .. math::\\alpha is the angle between k1 and k2

    Args:
        c_0,c_1,c_2 (:obj:`float`): Coefficients c_0,c_1 and c_2 of equation (1).

    Returns:
        (:obj:`tuple` of :obj:`float` or :obj:`None`): solutions if existing otherwise returns None.

    """

    if c_0==c_2:
        if c_1==0:
            t1 = t2 = None
        else:
            t1 = -c_0/c_1
            t2 = None
    else:
        rootval = (c_1/(c_2 - c_0))**2 -  (c_0 + c_2)/(c_2 - c_0)
        leadingterm = ( -c_1/(c_2 - c_0) ) 
        if rootval<0:
            t1,t2 = None, None
        else:
            t1 = leadingterm + np.sqrt( rootval )
            t2 = leadingterm - np.sqrt( rootval )

    s1, s2 = None, None

    if t1 is not None:
        s1 = 2 * np.arctan( t1 ) / alpha
        if s1>1 or s1<0:
            s1 = None

    if t2 is not None:
        s2 = 2 * np.arctan( t2 ) / alpha
        if s2>1 or s2<0:
            s2 = None
    
    return s1, s2


import numpy as np

class RigidBodyMotion(object):
    """Rigid body transform euclidean points by an euler axis and translation representation.

    The Motion is parametric in the interval time=[0,1] and will perform a rigid body transformation
    of a point x by uniformly rotating it from [0, rotation_angle] and translating [0, translation].
    I.e if called at a time time=t the motion will rotate the point t*rotation_angle radians and
    translate an amount t * translation.

    Args:
        rotation_axis (:obj:`numpy array`): Rotation axis ``shape=(3,)``
        rotation_angle (:obj:`float`): Radians for final rotation, when time=1.
        translation (:obj:`numpy array`):  Translation vector ``shape=(3,)``

    Attributes:
        rotation_axis (:obj:`numpy array`): Rotation axis ``shape=(3,)``
        rotation_angle (:obj:`float`): Radians for final rotation, when time=1.
        translation (:obj:`numpy array`):  Translation vector ``shape=(3,)``

    """

    def __init__(self, rotation_axis, rotation_angle, translation):
        assert rotation_angle < np.pi and  rotation_angle > 0, "The rotation angle must be in [0 pi]"
        self.rotator = _RodriguezRotator(rotation_axis)
        self.rotation_axis = rotation_axis
        self.rotation_angle = rotation_angle
        self.translation = translation

    def rotate( self, vectors, time ):
        """Find the rotational transformation of a set of vectors at a prescribed time.

        NOTE: This function only applies the rigid body rotation

        Args:
            vectors (:obj:`numpy array`): A set of points in 3d euclidean space to be rotated (``shape=(3,N)``)
            time (:obj:`float`): Time to compute for.

        Returns:
            Transformed vectors (:obj:`numpy array`) of ``shape=(3,N)``.

        """
        assert time <= 1 and time >= 0, "The rigid body motion is only valid on the interval time=[0,1]"
        return self.rotator( vectors, self.rotation_angle * time )

    def translate( self, vectors, time ):
        """Find the translational transformation of a set of points at a prescribed time.

        NOTE: This function only applies the rigid body translation

        Args:
            vectors (:obj:`numpy array`): A set of points in 3d euclidean space to be rotated (``shape=(3,N)``)
            time (:obj:`float`): Time to compute for.

        Returns:
            Transformed vectors (:obj:`numpy array`) of ``shape=(3,N)``.

        """
        assert time <= 1 and time >= 0, "The rigid body motion is only valid on the interval time=[0,1]"
        if len( vectors.shape )>1:
            translation = self.translation.reshape( 3, 1 )
        else:
            translation = self.translation
        return vectors + translation * time

    def __call__( self, vectors, time ):
        """Find the transformation of a set of points at a prescribed time.

        Args:
            vectors (:obj:`numpy array`): A set of points in 3d euclidean space to be rotated (``shape=(3,N)``)
            time (:obj:`float`): Time to compute for.

        Returns:
            Transformed vectors (:obj:`numpy array`) of ``shape=(3,N)``.

        """
        assert time <= 1 and time >= 0, "The rigid body motion is only valid on the interval time=[0,1]"
        if len( vectors.shape )>1:
            translation = self.translation.reshape( 3, 1 )
        else:
            translation = self.translation
        return self.rotator( vectors, self.rotation_angle * time ) + translation * time

class _RodriguezRotator(object):
    """Object for rotating vectors in the plane described by yhe unit normal rotation_axis.

    Args:
        rotation_axis (:obj:`numpy array`): A unit vector in 3d euclidean space (``shape=(3,)``)

    Attributes:
        rotation_axis (:obj:`numpy array`): A unit vector in 3d euclidean space (``shape=(3,)``)
        K (:obj:`numpy array`): (``shape=(3,3)``)
        K2 (:obj:`numpy array`): (``shape=(3,3)``)
        I (:obj:`numpy array`): (``shape=(3,3)``)

    """

    def __init__(self, rotation_axis):
        assert np.allclose( np.linalg.norm(rotation_axis), 1  ), "The rotation axis must be length unity."
        self.rotation_axis = rotation_axis
        rx,ry,rz = self.rotation_axis
        self.K = np.array([ [ 0, -rz,  ry],
                            [ rz,  0, -rx],
                            [ -ry, rx,  0] ])
        self.K2 = self.K.dot( self.K )
        self.I  = np.eye(3,3)

    def get_rotation_matrix(self, rotation_angle):
        return self.I + np.sin( rotation_angle )*self.K + ( 1 - np.cos( rotation_angle ) )*self.K2

    def __call__( self, vectors, rotation_angle ):
        """Rotate a vector in the plane described by v1 and v2 towards v2 a fraction s=[0,1].

        Args:
            vectors (:obj:`numpy array`): A set of vectors in 3d euclidean space to be rotated (``shape=(3,N)``)
            rotation_angle (:obj:`float`): Radians to rotate vectors around the rotation_axis (positive rotation).

        Returns:
            Rotated vectors (:obj:`numpy array`) of ``shape=(3,N)``.

        """
        R = self.get_rotation_matrix( rotation_angle )
        return R.dot( vectors )


import unittest
import numpy as np
from xfab import tools
from xrd_simulator import laue
from xrd_simulator import utils
from xrd_simulator.motion import RigidBodyMotion

class TestLaue(unittest.TestCase):

    """Unit test for the laue.py functions module. 
    """

    def setUp(self):
        np.random.seed(10) # changes all randomisation in the test

    def test_get_G_and_bragg_angle(self):
        U, B, cell, strain = self.get_pseudorandom_crystal()
        wavelength = self.get_pseudorandom_wavelength()
        
        G = laue.get_G(U, B, G_hkl=np.array([[1, 2, -1], [1, 3, -1]]).T)
        theta = laue.get_bragg_angle(G, wavelength)
        d = 2*np.pi / np.linalg.norm(G, axis=0)

        for i in range(len(theta)):        
            self.assertLessEqual( np.linalg.norm(G[:,i]), 4*np.pi / wavelength, msg="Length of G is wrong" )
            self.assertLessEqual( theta[i], np.pi, msg="Bragg angle too large" )
            self.assertGreaterEqual( theta[i], 0, msg="Bragg angle is negative" )
            self.assertAlmostEqual( np.sin(theta[i])*2*d[i], wavelength, msg="G and theta does not fulfill Braggs law" )

    def test_get_sin_theta_and_norm_G(self):
        U, B, cell, strain = self.get_pseudorandom_crystal()
        wavelength = self.get_pseudorandom_wavelength()
        G = laue.get_G(U, B, G_hkl=np.array([[1, 2, -1], [1, 3, -1]]).T)
        theta = laue.get_bragg_angle(G, wavelength)

        sinth, Gnorm = laue.get_sin_theta_and_norm_G( G, wavelength )

        for i in range(len(theta)):
            self.assertAlmostEqual(sinth[i], np.sin(theta[i]), msg="error theta")
            self.assertAlmostEqual(Gnorm[i], np.linalg.norm(G[:,i]), msg="error in norm of G")

    def test_get_tangens_half_angle_equation(self):
        wavelength = self.get_pseudorandom_wavelength()
        U, B, cell, strain = self.get_pseudorandom_crystal()
        k     = self.get_pseudorandom_wave_vector(wavelength)
        rotation_axis = np.random.rand(3,)
        rotation_axis = rotation_axis / np.linalg.norm( rotation_axis )
        rotation_angle = np.pi/2.
        G = laue.get_G(U, B, G_hkl=np.array([-1, 0, 0]))
        theta  = laue.get_bragg_angle( G, wavelength )
        c_0, c_1, c_2 = laue.get_tangens_half_angle_equation(k, theta, G, rotation_axis ) 
        self.assertTrue( np.isreal(c_0) )
        self.assertTrue( np.isreal(c_1) )
        self.assertTrue( np.isreal(c_2) )

    def test_find_solutions_to_tangens_half_angle_equation(self):
        U, B, cell, strain = self.get_pseudorandom_crystal()
        wavelength = cell[0]/18. # make sure the wavelength fits in the lattice spacing
        k = np.array([ 1.0, 0, 0 ])*2*np.pi/wavelength # select a large interval of k-vectors
        rotation_axis = np.random.rand(3,)
        rotation_axis = rotation_axis / np.linalg.norm( rotation_axis )
        rotation_angle = np.pi/2.
        U = np.eye(3,3) # let the crystal be aligned to assure 100 to be in its Bragg condition for the k-intervall
        G = laue.get_G(U, B, G_hkl=np.array([-1, 0, 0]))
        theta  = laue.get_bragg_angle( G, wavelength )
        c_0, c_1, c_2 = laue.get_tangens_half_angle_equation(k, theta, G, rotation_axis ) 
        t1, t2 = laue.find_solutions_to_tangens_half_angle_equation( c_0, c_1, c_2, rotation_angle )

        # Check that at least one solution has been found and that it satisfies the half angle equation.
        self.assertTrue( ( (t1 is not None) or (t2 is not None) ), msg="Tangens half angle equation could not be solved")
        if t1 is not None:
            self.assertLessEqual( t1, 1, msg="s>1")
            self.assertGreaterEqual( t1, 0, msg="s>1")
            t1 = np.tan( t1*rotation_angle/2. )
            self.assertAlmostEqual( (c_2 - c_0)*t1**2 + 2*c_1*t1 + (c_0 + c_2), 0, msg="Parametric solution wrong")
        if t2 is not None:
            self.assertLessEqual( t2, 1, msg="s>1")
            self.assertGreaterEqual( t2, 0, msg="s<0")
            t2 = np.tan( t1*rotation_angle/2. )
            self.assertAlmostEqual( (c_2 - c_0)*t2**2 + 2*c_1*t2 + (c_0 + c_2), 0, msg="Parametric solution wrong")

    def get_pseudorandom_crystal(self):
        phi1, PHI, phi2 = np.random.rand(3,)*2*np.pi
        U = tools.euler_to_u(phi1, PHI, phi2)
        strain_tensor = (np.random.rand(6,)-0.5)/50.
        unit_cell = [2+np.random.rand()*5, 2+np.random.rand()*5, 2+np.random.rand()*5, 90., 90., 90.]
        B = tools.epsilon_to_b( strain_tensor, unit_cell )
        return U, B, unit_cell, strain_tensor

    def get_pseudorandom_wavelength(self):
        return np.random.rand()*0.5

    def get_pseudorandom_wave_vector(self, wavelength):
        k = (np.random.rand(3,)-0.5)*2
        k = 2*np.pi*k/(np.linalg.norm(k)*wavelength)
        return k

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from xrd_simulator import beam
from xrd_simulator.beam import Beam
from scipy.spatial import ConvexHull
from xrd_simulator.motion import RigidBodyMotion

class TestBeam(unittest.TestCase):

    def setUp(self):
        np.random.seed(10) # changes all randomisation in the test
        self.beam_vertices = np.array([
            [-5., 0., 0. ],
            [-5., 1., 0. ],
            [-5., 0., 1. ],
            [-5., 1., 1. ],
            [ 5., 0., 0. ],
            [ 5., 1., 0. ],
            [ 5., 0., 1. ],
            [ 5., 1., 1. ]])
        self.wavelength = np.random.rand()*0.5
        self.xray_propagation_direction = np.random.rand(3,)
        self.xray_propagation_direction = self.xray_propagation_direction / np.linalg.norm(self.xray_propagation_direction)
        self.polarization_vector = np.random.rand(3,)
        self.polarization_vector = self.polarization_vector - np.dot( self.polarization_vector, self.xray_propagation_direction ) * self.xray_propagation_direction
        self.beam = Beam(self.beam_vertices, self.xray_propagation_direction, self.wavelength, self.polarization_vector)

    def test_intersect(self):
        vertices = self.beam_vertices
        ch = self.beam.intersect( vertices )
        for i in range(self.beam.vertices.shape[0]):
            v1 = ch.points[i,:]
            v2 = self.beam.vertices[i,:]
            self.assertAlmostEqual(  np.min( np.linalg.norm(v1 - self.beam.vertices, axis=1) ), 0 )

        vertices = self.beam_vertices
        vertices[:,0] = vertices[:,0]/2. 
        ch = self.beam.intersect( vertices )
        self.assertAlmostEqual(  ch.volume, 5 )

        ch1 = ConvexHull( ( np.random.rand(25,3)/2.+0.1 )  ) # polyhedra completely contained by the beam.
        vertices = ch1.points[ch1.vertices]
        ch = self.beam.intersect( vertices )
        self.assertAlmostEqual(  ch.volume, ch1.volume )

        ch1 = ConvexHull( ( np.random.rand(25,3)/2.+20.0 )  ) # polyhedra completely outside the beam.
        vertices = ch1.points[ch1.vertices]
        ch = self.beam.intersect( vertices )
        self.assertTrue(  ch is None )

    def test_get_proximity_intervals(self):

        self.beam_vertices = np.array([
            [-500., -1., -1. ],
            [-500., -1.,  1. ],
            [-500.,  1.,  1. ],
            [-500.,  1., -1. ],
            [500.,  -1., -1. ],
            [500.,  -1.,  1. ],
            [500.,   1.,  1. ],
            [500.,   1., -1. ]  ])
        self.beam_vertices[:,1:] = self.beam_vertices[:,1:]/10000000. # tiny beam cross section
        self.xray_propagation_direction = np.array([  1,  0,     0 ])
        self.polarization_vector = np.random.rand(3,)
        self.polarization_vector = self.polarization_vector - np.dot( self.polarization_vector, self.xray_propagation_direction ) * self.xray_propagation_direction
        self.beam = Beam(self.beam_vertices, self.xray_propagation_direction, self.wavelength, self.polarization_vector)

        rotation_axis  = np.array([ 0., 0., 1. ])
        rotation_angle = np.pi - 1e-8
        translation    = np.array([ 0., 0., 0. ])
        motion         = RigidBodyMotion(rotation_axis, rotation_angle, translation)

        sphere_centres = np.array([[400.0, 0.0, 0.0], [200.0, 0.0, 0.0]])
        sphere_radius = np.array([[2.0], [0.5]])

        intervals = self.beam.get_proximity_intervals(sphere_centres, sphere_radius, motion)
        
        self.assertEqual( len(intervals[0]), 2, msg="Wrong number of proximity intervals" )
        self.assertEqual( len(intervals[1]), 2, msg="Wrong number of proximity intervals" )

        for i in range(sphere_centres.shape[0]):
            # far away sphere small radii approximation:
            fraction_before_beam_leaves_sphere = np.arctan( sphere_radius[i] /  np.linalg.norm(sphere_centres[i]) ) / np.pi
            self.assertAlmostEqual( intervals[i][0][0], 0 , msg="Proximity interval wrong" )
            self.assertAlmostEqual( intervals[i][0][1], fraction_before_beam_leaves_sphere[0] , msg="Proximity interval wrong" )
            self.assertAlmostEqual( intervals[i][1][0], 1.-fraction_before_beam_leaves_sphere[0]  , msg="Proximity interval wrong" )
            self.assertAlmostEqual( intervals[i][1][1], 1.0 , msg="Proximity interval wrong" )

        motion.translation = np.array([ -87.24, 34.6 , 123.34 ]) # Now with rotation and translation

        intervals = self.beam.get_proximity_intervals(sphere_centres, sphere_radius, motion)
        
        self.assertEqual( len(intervals[0]), 1, msg="Wrong number of proximity intervals" )
        self.assertEqual( len(intervals[1]), 1, msg="Wrong number of proximity intervals" )

        for i in range(sphere_centres.shape[0]):
            # search numerically for the point in time when the sphere leaves the beam and compare to analytical intersection:
            times = np.linspace(0., 1., 10000)
            cs = np.array( [ motion(sphere_centres[i], t) for t in times ] )
            L = np.sqrt( cs[:,1]**2 + cs[:,2]**2 )
            fraction_before_beam_leaves_sphere = times[ np.argmin( np.abs(L-sphere_radius[i]) ) ]
            self.assertAlmostEqual( intervals[i][0][0], 0 , msg="Proximity interval wrong" )
            self.assertTrue( np.abs(intervals[i][0][1]-fraction_before_beam_leaves_sphere)<1./len(times) , msg="Proximity interval wrong" )

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from xrd_simulator.polycrystal import Polycrystal
from xrd_simulator.mesh import TetraMesh
from xrd_simulator.phase import Phase
from xrd_simulator.detector import Detector
from xrd_simulator.beam import Beam
from xrd_simulator.motion import RigidBodyMotion
from xfab import tools
from scipy.signal import convolve
import os

class TestPolycrystal(unittest.TestCase):

    def setUp(self):
        np.random.seed(10)

        self.pixel_size = 75.
        self.detector_size = self.pixel_size*1024
        self.detector_distance = 142938.28756189224
        self.d0 = np.array([self.detector_distance,   -self.detector_size/2.,  -self.detector_size/2.])
        self.d1 = np.array([self.detector_distance,    self.detector_size/2.,  -self.detector_size/2.])
        self.d2 = np.array([self.detector_distance,   -self.detector_size/2.,   self.detector_size/2.])

        self.detector = Detector( self.pixel_size, self.pixel_size, self.d0, self.d1, self.d2 )

        mesh = TetraMesh.generate_mesh_from_levelset(
            level_set = lambda x: np.dot( x, x ) - self.detector_size/10.,
            bounding_radius = 1.1*self.detector_size/10., 
            max_cell_circumradius = 0.01*self.detector_size/10. )

        unit_cell = [4.926, 4.926, 5.4189, 90., 90., 120.]
        sgname = 'P3221' # Quartz
        phases = [Phase(unit_cell, sgname)]
        B0 = tools.epsilon_to_b( np.zeros((6,)), unit_cell )
        eB = np.array( [ B0 for _ in range(mesh.number_of_elements)] )
        euler_angles = np.random.rand(mesh.number_of_elements, 3) * 2 * np.pi
        eU = np.array( [tools.euler_to_u(ea[0], ea[1], ea[2]) for ea in euler_angles] )
        ephase = np.zeros((mesh.number_of_elements,)).astype(int)
        self.polycrystal = Polycrystal(mesh, ephase, eU, eB, phases)

        w = self.detector_size/2. # full field illumination
        beam_vertices = np.array([
            [-self.detector_distance, -w, -w ],
            [-self.detector_distance,  w, -w ],
            [-self.detector_distance,  w,  w ],
            [-self.detector_distance, -w,  w ],
            [ self.detector_distance, -w, -w  ],
            [ self.detector_distance,  w, -w  ],
            [ self.detector_distance,  w,  w  ],
            [ self.detector_distance, -w,  w  ]])
        wavelength = 0.285227
        xray_propagation_direction = np.array([1,0,0]) * 2 * np.pi / wavelength
        polarization_vector = np.array([0,1,0])
        self.beam = Beam(beam_vertices, xray_propagation_direction, wavelength, polarization_vector)

    def test_diffract(self):

        rotation_angle = 10*np.pi/180.
        rotation_axis = np.array([0,0,1])
        translation = np.array([0,0,0])
        motion  = RigidBodyMotion(rotation_axis, rotation_angle, translation)

        self.polycrystal.diffract( self.beam, self.detector, motion )

        # The rendered diffraction pattern should have intensity
        diffraction_pattern = self.detector.render(frame_number=0, lorentz=True, polarization=True, structure_factor=True)

        self.assertGreater(np.sum(diffraction_pattern), 0)

        # .. and the intensity should be scattered over the image
        w = int(self.detector_size/5.)
        for i in range(w, diffraction_pattern.shape[0]-w, w ):
            for j in range(w, diffraction_pattern.shape[0]-w, w ):
                subsum = np.sum( diffraction_pattern[i-w:i+w,j-w:j+w] )
                self.assertGreaterEqual(subsum, 0)

        # Scatterers should be confined to rings
        bragg_angles = []
        for scatterer in self.detector.frames[0]:
            kprime = scatterer.scattered_wave_vector
            k = scatterer.incident_wave_vector
            normfactor = np.linalg.norm(k)*np.linalg.norm(kprime)
            tth = np.arccos( np.dot(k, kprime)/normfactor )
            bragg_angles.append(tth/2.)

        # count the number of non-overlaing rings there should be quite a few.
        bragg_angles =  np.degrees( bragg_angles )
        hist = np.histogram(bragg_angles, bins=np.linspace(0,20,360))[0]
        csequence, nosequences = 0, 0
        for i in range(hist.shape[0]):
            if hist[i]>0:
                csequence +=1
            elif csequence>=1:
                    nosequences +=1
                    csequence = 0
        self.assertGreaterEqual(nosequences, 20, msg="Few or no rings appeared from diffraction.")

    def test_save_and_load(self):
        eB = self.polycrystal.eB.copy()
        path = os.path.join( os.path.join(os.path.dirname(__file__), 'data' ), 'my_polycrystal' )
        self.polycrystal.save( path )
        self.polycrystal  = Polycrystal.load( path )
        self.assertTrue( np.allclose( eB, self.polycrystal.eB ), msg='Data corrupted on save and load' )
        os.remove( path )

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from xrd_simulator.scatterer import Scatterer
from scipy.spatial import ConvexHull
import pkg_resources
from xrd_simulator.phase import Phase
import os

class TestScatterer(unittest.TestCase):

    def setUp(self):
        np.random.seed(10)
        self.wavelength = 1.0
        verts = np.array([[0,0,0],
                          [0,0,1],
                          [0,1,0],
                          [1,0,0],
                          [0.1 ,0.1, 0.1]])
        self.ch = ConvexHull( verts )
        scattered_wave_vector = np.random.rand(3,)
        self.scattered_wave_vector = 2*np.pi*scattered_wave_vector/(self.wavelength * np.linalg.norm(scattered_wave_vector) )
        self.time = np.random.rand()

        data = os.path.join( os.path.join(os.path.dirname(__file__), 'data' ), 'Fe_mp-150_conventional_standard.cif' )
        unit_cell = [3.64570000, 3.64570000, 3.64570000, 90.0, 90.0, 90.0]
        sgname = 'Fm-3m' # Iron
        self.phase   = Phase(unit_cell, sgname, path_to_cif_file=data)
        self.phase.setup_diffracting_planes(self.wavelength, 0, 20*np.pi/180)

        self.incident_wave_vector = np.array([1,0.,0])
        self.incident_wave_vector = 2*np.pi*self.incident_wave_vector/(self.wavelength * np.linalg.norm(self.incident_wave_vector) )
        self.incident_polarization_vector = np.array([0.,1.,0]) 
        self.rotation_axis = np.array([0.,0, 1.])

        self.scatterer = Scatterer( self.ch, 
                                    self.scattered_wave_vector, 
                                    self.incident_wave_vector, 
                                    self.wavelength,
                                    self.incident_polarization_vector, 
                                    self.rotation_axis,
                                    self.time, 
                                    self.phase, 
                                    hkl_indx=0 )

    def test_lorentz(self):
        z = np.array([0,1,1.])
        self.scatterer.scattered_wave_vector = 2*np.pi*z/(np.linalg.norm(z)*self.wavelength)
        L = self.scatterer.lorentz_factor
        self.assertAlmostEqual( L, np.sqrt(2.) )  

        z = np.array([0,0,1.])
        self.scatterer.scattered_wave_vector = 2*np.pi*z/(self.wavelength)
        L = self.scatterer.lorentz_factor
        self.assertTrue( L is np.inf )

    def test_polarization(self):
        self.scatterer.scattered_wave_vector = np.array([0, 1.0, 0])
        P = self.scatterer.polarization_factor
        self.assertAlmostEqual( P, 0 )  

        self.scatterer.scattered_wave_vector = np.array([1./np.sqrt(2), 0, -1./np.sqrt(2)])
        P = self.scatterer.polarization_factor
        self.assertAlmostEqual( P, 1.0 ) 

    def test_hkl(self):
        hkl = self.scatterer.hkl
        for i in range(3):
            self.assertAlmostEqual( hkl[i], -1, msg="hkl is wrong" )

    def test_structure_factor(self):
        structure_factor = self.scatterer.real_structure_factor
        self.assertGreaterEqual( structure_factor, 0, msg="structure factor is wrong" )

    def test_centroid(self):
        centroid2 = self.scatterer.centroid
        for c1,c2 in zip(np.array([0.25, 0.25, 0.25]), centroid2):
            self.assertAlmostEqual( c1, c2, msg="centroid is wrong" )

    def test_volume(self):
        vol = self.scatterer.volume
        self.assertAlmostEqual( vol, 1/6., msg="volume is wrong" )

if __name__ == '__main__':
    unittest.main()


import numpy as np
import pygalmesh
from scipy.spatial.transform import Rotation
from xrd_simulator.detector import Detector
from xrd_simulator.beam import Beam
from xrd_simulator.motion import RigidBodyMotion
from xrd_simulator.polycrystal import Polycrystal
from xrd_simulator.mesh import TetraMesh
from xrd_simulator.phase import Phase
from xrd_simulator import utils
from xfab import tools

PARAMETER_KEYS = [
    "detector_distance",
    "number_of_detector_pixels_z",
    "number_of_detector_pixels_y",
    "detector_center_pixel_z",
    "detector_center_pixel_y",
    "pixel_side_length_z",
    "pixel_side_length_y",
    "wavelength",
    "beam_side_length_z",
    "beam_side_length_y",
    "rotation_step",
    "rotation_axis"
]


def s3dxrd(parameters):
    """Construct a scaning-three-dimensional-xray diffraction experiment.

    This is a helper/utility function for quickly creating an experiment. For full controll
    over the diffraction geometry consider custom creation of the primitive quantities:
    (:obj:`xrd_simulator.beam.Beam`), and (:obj:`xrd_simulator.detector.Detector`) seperately.

    Args:
        parameters (:obj:`dict`): Dictionary with fields as \n
            ``"detector_distance"``           : (:obj:`float`) Distance form sample origin to
                detector centre in units of microns. \n
            ``"number_of_detector_pixels_z"`` : (:obj:`int`) Number of detector pixels
                along z-axis. \n
            ``"number_of_detector_pixels_y"`` : (:obj:`int`) Number of detector pixels
                along y-axis. \n
            ``"detector_center_pixel_z"``     : (:obj:`float`) Intersection pixel coordinate between
                 beam centroid line and detector along z-axis. \n
            ``"detector_center_pixel_y"``     : (:obj:`float`) Intersection pixel coordinate between
                 beam centroid line and detector along y-axis. \n
            ``"pixel_side_length_z"``         : (:obj:`float`) Detector pixel side length in units
                of microns along z-axis. \n
            ``"pixel_side_length_y"``         : (:obj:`float`) Detector pixel side length in units
                of microns along y-axis. \n
            ``"wavelength"``                  : (:obj:`float`) Wavelength in units of Angstrom. \n
            ``"beam_side_length_z"``          : (:obj:`float`) Beam side length in units
                of microns. \n
            ``"beam_side_length_y"``          : (:obj:`float`) Beam side length in units
                of microns. \n
            ``"rotation_step"``               : (:obj:`float`) Angular frame integration step in
                units of radians. \n
            ``"rotation_axis"``               : (:obj:`numpy array`)  Axis around which to
                positively rotate the sample by ``rotation_step`` radians. \n

    Returns:
        (:obj:`xrd_simulator`) objects defining an experiment: (:obj:`xrd_simulator.beam.Beam`),
        (:obj:`xrd_simulator.detector.Detector`), (:obj:`xrd_simulator.motion.RigidBodyMotion`).

    Examples:
        .. literalinclude:: examples/s3dxrd.py

    """
    for key in PARAMETER_KEYS:
        if key not in list(parameters):
            raise ValueError("No keyword " + key + " found in the input parameters dictionary")

    detector = _get_detector_from_params(parameters)
    beam = _get_beam_from_params(parameters)
    motion = _get_motion_from_params(parameters)

    return beam, detector, motion


def _get_motion_from_params(parameters):
    """Produce a ``xrd_simulator.motion.RigidBodyMotion`` from the s3dxrd params dictionary.
    """
    translation = np.array([0., 0., 0.])
    return RigidBodyMotion(parameters["rotation_axis"], parameters["rotation_step"], translation)


def _get_beam_from_params(parameters):
    """Produce a ``xrd_simulator.beam.Beam`` from the s3dxrd params dictionary.
    """
    dz = parameters['beam_side_length_z'] / 2.
    dy = parameters['beam_side_length_y'] / 2.
    beam_vertices = np.array([
        [-parameters['detector_distance'], -dy, -dz],
        [-parameters['detector_distance'], dy, -dz],
        [-parameters['detector_distance'], -dy, dz],
        [-parameters['detector_distance'], dy, dz],
        [parameters['detector_distance'], -dy, -dz],
        [parameters['detector_distance'], dy, -dz],
        [parameters['detector_distance'], -dy, dz],
        [parameters['detector_distance'], dy, dz]
    ])

    beam_direction = np.array([1.0, 0.0, 0.0])
    polarization_vector = np.array([0.0, 1.0, 0.0])

    return Beam(beam_vertices, beam_direction, parameters['wavelength'], polarization_vector)


def _get_detector_from_params(parameters):
    """Produce a ``xrd_simulator.detector.Detector`` from the s3dxrd params dictionary.
    """

    p_z, p_y = parameters['pixel_side_length_z'], parameters['pixel_side_length_y']
    det_dist = parameters['detector_distance']
    det_pix_y = parameters['number_of_detector_pixels_y']
    det_pix_z = parameters['number_of_detector_pixels_z']
    det_cz, det_cy = parameters['detector_center_pixel_z'], parameters['detector_center_pixel_y']

    d_0 = np.array([det_dist, -det_cy * p_y, -det_cz * p_z])
    d_1 = np.array([det_dist, (det_pix_y - det_cy) * p_y, -det_cz * p_z])
    d_2 = np.array([det_dist, -det_cy * p_y, (det_pix_z - det_cz) * p_z])

    return Detector(p_z, p_y, d_0, d_1, d_2)


def polycrystal_from_odf(orientation_density_function,
                         number_of_crystals,
                         sample_bounding_cylinder_height,
                         sample_bounding_cylinder_radius,
                         unit_cell,
                         sgname,
                         maximum_sampling_bin_seperation=np.radians(5.0),
                         strain_tensor=lambda x: np.zeros((3, 3))):
    """Fill a cylinder with crystals from a given orientation density function.

    The ``orientation_density_function`` is sampled by discretizing orientation space over the unit
    quarternions. Each bin is assigned its apropiate probability, assuming the
    ``orientation_density_function`` is approximately constant over a single bin. Each sampled
    orientation is constructed by first drawing a random bin and next drawing uniformly from
    within that bin, again assuming that ``orientation_density_function`` is approximately constant
    over a bin.

    Args:
        orientation_density_function (:obj:`callable`): orientation_density_function(x, q) ->
            :obj:`float` where input variable ``x`` is a :obj:`numpy array` of shape ``(3,)``
            representing a spatial coordinate in the cylinder (x,y,z) and ``q`` is a
            :obj:`numpy array` of shape ``(4,)`` representing a orientation in so3 by a unit
            quarternion. The format of the quarternion is "scalar last"
            (same as in scipy.spatial.transform.Rotation).
        number_of_crystals (:obj:`int`): Approximate number of crystal elements to compose the
            cylinder volume.
        sample_bounding_cylinder_height (:obj:`float`): Height of sample cylinder in units of
             microns.
        sample_bounding_cylinder_radius (:obj:`float`): Radius of sample cylinder in units of
            microns.
        unit_cell (:obj:`list` of :obj:`float`): Crystal unit cell representation of the form
            [a,b,c,alpha,beta,gamma], where alpha,beta and gamma are in units of degrees while
            a,b and c are in units of anstrom.
        sgname (:obj:`string`): Name of space group , e.g 'P3221' for quartz, SiO2, for instance
        maximum_sampling_bin_seperation (:obj:`float`): Discretization steplength of orientation
            space using spherical coordinates over the unit quarternions in units of radians.
            A smaller steplength gives more accurate sampling of the input
            ``orientation_density_function`` but is computationally slower.
        strain_tensor (:obj:`callable`): Strain tensor field over sample cylinder.
            strain_tensor(x) -> :obj:`numpy array` of shape ``(3,3)`` where input variable ``x`` is
            a :obj:`numpy array` of shape ``(3,)`` representing a spatial coordinate in the
            cylinder (x,y,z).

    Returns:
        (:obj:`xrd_simulator.polycrystal.Polycrystal`)

    Examples:
        .. literalinclude:: examples/polycrystal_from_odf.py

    """
    # Sample topology
    volume_per_crystal = np.pi * (sample_bounding_cylinder_radius**2) * \
        sample_bounding_cylinder_height / number_of_crystals
    max_cell_circumradius = (3 * volume_per_crystal / (np.pi * 4.))**(1 / 3.)

    # Fudge factor 2.6 gives approximately number_of_crystals elements in the mesh
    max_cell_circumradius = 2.65 * max_cell_circumradius

    dz = sample_bounding_cylinder_height / 2.
    R = float(sample_bounding_cylinder_radius)

    cylinder = pygalmesh.generate_mesh(
        pygalmesh.Cylinder(-dz, dz, R, max_cell_circumradius),
        max_cell_circumradius=max_cell_circumradius,
        max_edge_size_at_feature_edges=max_cell_circumradius,
        verbose=False)

    mesh = TetraMesh._build_tetramesh(cylinder)

    # Sample is uniformly single phase
    phases = [Phase(unit_cell, sgname)]
    ephase = np.zeros((mesh.number_of_elements,)).astype(int)

    # Sample spatial texture
    eU = _sample_ODF(orientation_density_function, maximum_sampling_bin_seperation, mesh.ecentroids)

    # Sample spatial strain
    eB = np.zeros((mesh.number_of_elements, 3, 3))
    for ei in range(mesh.number_of_elements):
        strain_lab = strain_tensor(mesh.ecentroids[ei])  # strain in lab-coordinates
        eB[ei] = utils.lab_strain_to_lattice_matrix(strain_lab, eU[ei], unit_cell)

    return Polycrystal(mesh, ephase, eU, eB, phases)


def _sample_ODF(ODF, maximum_sampling_bin_seperation, coordinates):
    """Draw orientation matrices form an ODF at spatial locations ``coordinates``.
    """

    dalpha = maximum_sampling_bin_seperation / 2.  # TODO: verify this analytically.
    dalpha = np.pi / 2. / int(np.pi / (dalpha * 2.))
    alpha_1 = np.arange(0 + dalpha / 2. + 1e-8, np.pi / 2. - dalpha / 2. - 1e-8 + dalpha, dalpha)
    alpha_2 = np.arange(0 + dalpha / 2. + 1e-8, np.pi - dalpha / 2. - 1e-8 + dalpha, dalpha)
    alpha_3 = np.arange(0 + dalpha / 2. + 1e-8, 2 * np.pi - dalpha / 2. - 1e-8 + dalpha, dalpha)

    A1, A2, A3 = np.meshgrid(alpha_1, alpha_2, alpha_3, indexing='ij')
    A1, A2, A3 = A1.flatten(), A2.flatten(), A3.flatten()

    q = utils.alpha_to_quarternion(A1, A2, A3)

    # Approximate volume ber bin:
    # volume_element = (np.sin(A1)**2) * np.sin(A2) * (dalpha**3)

    # Actual volume per bin requires integration as below:
    da = dalpha / 2.
    a = ((1 / 2.) * (A1 + da - np.sin(A1 + da) * np.cos(A1 + da)) -
         (1 / 2.) * (A1 - da - np.sin(A1 - da) * np.cos(A1 - da)))
    b = (-np.cos(A2 + da) + np.cos(A2 - da))
    c = ((A3 + da) - (A3 - da))
    volume_element = a * b * c
    assert np.abs(np.sum(volume_element) - (np.pi**2)) < 1e-5

    rotations = []
    for x in coordinates:
        probability = volume_element * ODF(x, q)
        assert np.abs(np.sum(probability) -
                      1.0) < 0.05, "Orientation density function must be be normalised."
        # Normalisation is not exact due to the discretization.
        probability = probability / np.sum(probability)
        indices = np.linspace(0, len(probability) - 1, len(probability)).astype(int)
        draw = np.random.choice(indices, size=1, replace=True, p=probability)
        a1 = A1[draw] + dalpha * (np.random.rand() - 0.5)
        a2 = A2[draw] + dalpha * (np.random.rand() - 0.5)
        a3 = A3[draw] + dalpha * (np.random.rand() - 0.5)
        q_pertubated = utils.alpha_to_quarternion(a1, a2, a3)
        rotations.append(Rotation.from_quat(q_pertubated).as_matrix().reshape(3, 3))

    return np.array(rotations)


def get_uniform_powder_sample(
        sample_bounding_radius,
        number_of_grains,
        unit_cell,
        sgname,
        strain_tensor=np.zeros((3, 3))):
    """Generate a polycyrystal with grains overlayed at the origin and orientations drawn uniformly.

    Args:
        sample_bounding_radius (:obj:`float`): Bounding radius of sample. All tetrahedral crystal
            elements will be overlayed within a sphere of ``sample_bounding_radius`` radius.
        number_of_grains (:obj:`int`): Number of grains composing the polycrystal sample.
        unit_cell (:obj:`list` of :obj:`float`): Crystal unit cell representation of the form
            [a,b,c,alpha,beta,gamma], where alpha,beta and gamma are in units of degrees while
            a,b and c are in units of anstrom.
        sgname (:obj:`string`):  Name of space group , e.g 'P3221' for quartz, SiO2, for instance
        strain_tensor (:obj:`numpy array`): Strain tensor to apply to all tetrahedral crystal
            elements contained within the sample. ``shape=(3,3)``.

    Returns:
        (:obj:`xrd_simulator.polycrystal`) A polycyrystal sample with ``number_of_grains`` grains.

    Examples:
        .. literalinclude:: examples/get_uniform_powder_sample.py

    """
    coord, enod, eU, node_number = [], [], [], 0
    r = sample_bounding_radius
    for _ in range(number_of_grains):
        coord.append([r / np.sqrt(3.), r / np.sqrt(3.), -r / np.sqrt(3.)])
        coord.append([r / np.sqrt(3.), -r / np.sqrt(3.), -r / np.sqrt(3.)])
        coord.append([-r / np.sqrt(2.), 0, -r / np.sqrt(2.)])
        coord.append([0, 0, r])
        enod.append(list(range(node_number, node_number + 4)))
        node_number += 3
    coord, enod = np.array(coord), np.array(enod)
    mesh = TetraMesh.generate_mesh_from_vertices(coord, enod)
    eU = Rotation.random(mesh.number_of_elements).as_matrix()
    eB = np.array([utils.lab_strain_to_lattice_matrix(strain_tensor, U, unit_cell) for U in eU])
    ephase = np.zeros((mesh.number_of_elements,)).astype(int)
    return Polycrystal(mesh, ephase, eU, eB, [Phase(unit_cell, sgname)])



import unittest
import numpy as np
from xfab import tools
from xrd_simulator import utils
from scipy.spatial.transform import Rotation

class TestUtils(unittest.TestCase):

    def setUp(self):
        np.random.seed(10) # changes all randomisation in the test

    def test_clip_line_with_convex_polyhedron(self):
        line_points    = np.ascontiguousarray( [[-1.,0.2,0.2],[-1.,0.4,0.6]] )
        line_direction = np.ascontiguousarray( [1.0, 0.0, 0.0] )
        line_direction = line_direction/np.linalg.norm(line_direction)
        plane_points   = np.ascontiguousarray( [[0.,0.5,0.5], [1,0.5,0.5], [0.5,0.5,0.], [0.5,0.5,1.], [0.5,0,0.5], [0.5,1.,0.5]] )
        plane_normals  = np.ascontiguousarray( [[-1.,0.,0.],[1.,0.,0.], [0.,0.,-1.],[0.,0.,1.], [0.,-1.,0.],[0.,1.,0.]] )
        clip_lengths   = utils.clip_line_with_convex_polyhedron( line_points, line_direction, plane_points, plane_normals )
        for l in clip_lengths: 
            self.assertAlmostEqual(l, 1.0, msg="Projection through unity cube should give unity clip length")

        line_direction = np.ascontiguousarray( [1.0, 0.2, 0.1] )
        line_direction = line_direction/np.linalg.norm(line_direction)
        clip_lengths   = utils.clip_line_with_convex_polyhedron( line_points, line_direction, plane_points, plane_normals )
        for l in clip_lengths: 
            self.assertGreater(l, 1.0, msg="Titlted projection through unity cube should give greater than unity clip length")

    def test_lab_strain_to_lattice_matrix(self):

        U = Rotation.random().as_matrix()
        strain_tensor = (np.random.rand(3,3)-0.5)/100. # random small strain tensor
        strain_tensor = (strain_tensor.T + strain_tensor)/2.
        unit_cell = [5.028, 5.028, 5.519, 90., 90., 120.]
        B = utils.lab_strain_to_lattice_matrix(strain_tensor, U, unit_cell)

        n_c = np.random.rand(3,) # crystal unit vector
        n_c = n_c / np.linalg.norm(n_c)
        n_l = np.dot( U, n_c ) # lab unit vector

        # strain along n_l described in lab frame
        strain_l = np.dot( np.dot( n_l, strain_tensor ), n_l) 
        s = tools.b_to_epsilon( B, unit_cell )
        crystal_strain = np.array([[s[0],s[1],s[2]],[s[1],s[3],s[4]],[s[2],s[4],s[5]]])

        # strain along n_l described in crystal frame
        strain_c = np.dot( np.dot( n_c, crystal_strain ), n_c) 

        # The strain should be invariant along a direction
        self.assertAlmostEqual(strain_l, strain_c, msg="bad crystal to lab frame conversion")

    def test_alpha_to_quarternion(self):
        _, alpha_2, alpha_3 = np.random.rand(3,)
        q = utils.alpha_to_quarternion(0, alpha_2, alpha_3)
        self.assertAlmostEqual(q[0], 1.0, msg="quarternion wrongly computed")
        self.assertAlmostEqual(q[1], 0.0, msg="quarternion wrongly computed")
        self.assertAlmostEqual(q[2], 0.0, msg="quarternion wrongly computed")
        self.assertAlmostEqual(q[3], 0.0, msg="quarternion wrongly computed")
        alpha_1 = np.random.rand(7,)
        alpha_2 = np.random.rand(7,)
        alpha_3 = np.random.rand(7,)
        qq = utils.alpha_to_quarternion(alpha_1, alpha_2, alpha_3)
        for q in qq:
            self.assertTrue(np.abs(np.linalg.norm(q)-1.0)<1e-5, msg="quarternion not normalised")

    def test_diffractogram(self):
        diffraction_pattern = np.zeros((20,20))
        R = 8
        det_c_z, det_c_y = 10.,10.
        for i in range(diffraction_pattern.shape[0]):
            for j in range(diffraction_pattern.shape[1]):
                if np.abs( np.sqrt( (i-det_c_z)**2 + (j-det_c_y)**2 ) - R) < 0.5:
                    diffraction_pattern[i,j]+=1
        bin_centres, histogram = utils.diffractogram(diffraction_pattern, det_c_z, det_c_y, 1.0)
        self.assertEqual(np.sum(histogram>0), 1, msg="Error in diffractogram azimuth integration" )
        self.assertEqual(np.sum(histogram), np.sum(diffraction_pattern), msg="Error in diffractogram azimuth integration" )
        self.assertEqual(histogram[R], np.sum(diffraction_pattern), msg="Error in diffractogram azimuth integration" )

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from scipy.spatial.transform import Rotation
from xrd_simulator import templates, utils
import matplotlib.pyplot as plt

class TestUtils(unittest.TestCase):

    def setUp(self):
        np.random.seed(5) # changes all randomization in the test

    def test_s3dxrd(self):

        parameters = {
            "detector_distance"             : 191023.9164,
            "detector_center_pixel_z"       : 256.2345,
            "detector_center_pixel_y"       : 255.1129,
            "pixel_side_length_z"           : 181.4234,
            "pixel_side_length_y"           : 180.2343,
            "number_of_detector_pixels_z"   : 512,
            "number_of_detector_pixels_y"   : 512,
            "wavelength"                    : 0.285227,
            "beam_side_length_z"            : 512 * 200.,
            "beam_side_length_y"            : 512 * 200.,
            "rotation_step"                 : np.radians(1.634),
            "rotation_axis"                 : np.array([0., 0., 1.0])
        }

        beam, detector, motion = templates.s3dxrd( parameters )
        for ci in beam.centroid:
            self.assertAlmostEqual(ci, 0, msg="beam not at origin.")

        det_approx_centroid = detector.d0.copy()
        det_approx_centroid[1] += detector.d1[1]
        det_approx_centroid[2] += detector.d2[2]

        self.assertAlmostEqual(det_approx_centroid[0], parameters["detector_distance"], msg="Detector distance wrong.")
        self.assertLessEqual(np.abs(det_approx_centroid[1]), 5*parameters["pixel_side_length_y"], msg="Detector not centered.")
        self.assertLessEqual(np.abs(det_approx_centroid[2]), 5*parameters["pixel_side_length_z"], msg="Detector not centered.")

        original_vector = np.random.rand(3,)-0.5
        time = 0.234986
        transformed_vector = motion( original_vector, time )

        angle = parameters["rotation_step"] * time
        s,c = np.sin(angle),np.cos(angle)
        Rz = np.array([[c,-s,0],[s,c,0],[0,0,1]])

        self.assertAlmostEqual(transformed_vector[0], np.dot(Rz,original_vector)[0], msg="Motion does not rotate around z-axis")
        self.assertAlmostEqual(transformed_vector[1], np.dot(Rz,original_vector)[1], msg="Motion does not rotate around z-axis")
        self.assertAlmostEqual(transformed_vector[2], original_vector[2], msg="Motion does not rotate around z-axis")


    def test_polycrystal_from_odf(self):

        unit_cell = [4.926, 4.926, 5.4189, 90., 90., 120.]
        sgname = 'P3221' # Quartz
        orientation_density_function = lambda x,q: 1./(np.pi**2) # uniform ODF
        number_of_crystals = 500
        sample_bounding_cylinder_height = 50
        sample_bounding_cylinder_radius = 25
        maximum_sampling_bin_seperation = np.radians(10.0)
        strain_tensor = lambda x: np.array([[0,0, 0.02*x[2]/sample_bounding_cylinder_height],[0,0,0],[0,0,0]]) # Linear strain gradient along rotation axis.

        polycrystal = templates.polycrystal_from_odf(  orientation_density_function,
                                                       number_of_crystals,
                                                       sample_bounding_cylinder_height,
                                                       sample_bounding_cylinder_radius,
                                                       unit_cell,
                                                       sgname,
                                                       maximum_sampling_bin_seperation,
                                                       strain_tensor )

        # Compare Euler angle distributions to scipy random uniform orientation sampler
        euler1 = np.array([ Rotation.from_matrix( U ).as_euler( 'xyz', degrees=True ) for U in polycrystal.eU_lab])
        euler2 = Rotation.random( 10 * euler1.shape[0] ).as_euler('xyz')

        for i in range(3):
            hist1, bins = np.histogram(euler1[:,i])
            hist2, bins = np.histogram(euler2[:,i])
            hist2 = hist2 / 10.
            # These histograms should look roughly the same
            self.assertLessEqual( np.max(np.abs(hist1 - hist2)), number_of_crystals*0.05, "ODF not sampled correctly." )

        parameters = {
            "detector_distance"             : 191023.9164,
            "detector_center_pixel_z"       : 256.2345,
            "detector_center_pixel_y"       : 255.1129,
            "pixel_side_length_z"           : 181.4234,
            "pixel_side_length_y"           : 180.2343,
            "number_of_detector_pixels_z"   : 512,
            "number_of_detector_pixels_y"   : 512,
            "wavelength"                    : 0.285227,
            "beam_side_length_z"            : 512 * 200.,
            "beam_side_length_y"            : 512 * 200.,
            "rotation_step"                 : np.radians(20.0),
            "rotation_axis"                 : np.array([0., 0., 1.0])
        }

        beam, detector, motion = templates.s3dxrd( parameters )

        number_of_crystals = 100
        sample_bounding_cylinder_height = 256 * 180 / 128.
        sample_bounding_cylinder_radius = 256 * 180 / 128.

        polycrystal = templates.polycrystal_from_odf( orientation_density_function,
                                                        number_of_crystals,
                                                        sample_bounding_cylinder_height,
                                                        sample_bounding_cylinder_radius,
                                                        unit_cell,
                                                        sgname,
                                                        maximum_sampling_bin_seperation,
                                                        strain_tensor )

        polycrystal.transform(motion, time=0.134)
        polycrystal.diffract( beam, detector, motion, min_bragg_angle=0, max_bragg_angle=None, verbose=True )
        diffraction_pattern = detector.render(frame_number=0,lorentz=False, polarization=False, structure_factor=False, method="centroid", verbose=True)
        bins, histogram = utils.diffractogram(diffraction_pattern>1, parameters['detector_center_pixel_z'], parameters['detector_center_pixel_y'])

        csequence, nosequences = 0, 0
        for i in range(histogram.shape[0]):
            if histogram[i]>0:
                csequence +=1
            elif csequence>=1:
                    nosequences +=1
                    csequence = 0
        self.assertGreaterEqual(nosequences, 10, msg="Few or no rings appeared from diffraction.")


    def test_get_uniform_powder_sample(self):
        sample_bounding_radius = 256 * 180 / 128.
        polycrystal = templates.get_uniform_powder_sample(
                        sample_bounding_radius,
                        number_of_grains = 50,
                        unit_cell = [4.926, 4.926, 5.4189, 90., 90., 120.],
                        sgname = 'P3221',
                        strain_tensor = np.array([[0,0,0],[0,0,0],[0,0,0.01]])
                        )
        for c in polycrystal.mesh_lab.coord:
            self.assertLessEqual( np.linalg.norm(c), sample_bounding_radius+1e-8, msg='Powder sample not contained by bounding sphere.' )

        parameters = {
            "detector_distance"             : 191023.9164,
            "detector_center_pixel_z"       : 256.2345,
            "detector_center_pixel_y"       : 255.1129,
            "pixel_side_length_z"           : 181.4234,
            "pixel_side_length_y"           : 180.2343,
            "number_of_detector_pixels_z"   : 512,
            "number_of_detector_pixels_y"   : 512,
            "wavelength"                    : 0.285227,
            "beam_side_length_z"            : 512 * 200.,
            "beam_side_length_y"            : 512 * 200.,
            "rotation_step"                 : np.radians(20.0),
            "rotation_axis"                 : np.array([0., 0., 1.0])
        }

        beam, detector, motion = templates.s3dxrd( parameters )
        polycrystal.transform(motion, time=0.234)

        polycrystal.diffract( beam, detector, motion, min_bragg_angle=0, max_bragg_angle=None, verbose=True )
        diffraction_pattern = detector.render(frame_number=0,lorentz=False, polarization=False, structure_factor=False, method="centroid", verbose=True)
        bins, histogram = utils.diffractogram(diffraction_pattern>1, parameters['detector_center_pixel_z'], parameters['detector_center_pixel_y'])

        csequence, nosequences = 0, 0
        for i in range(histogram.shape[0]):
            if histogram[i]>0:
                csequence +=1
            elif csequence>=1:
                    nosequences +=1
                    csequence = 0
        self.assertGreaterEqual(nosequences, 10, msg="Few or no rings appeared from diffraction.")

if __name__ == '__main__':
    unittest.main()


import numpy as np 
import matplotlib.pyplot as plt
from numpy.lib.utils import source
from xrd_simulator.scatterer import Scatterer
from xrd_simulator import utils, laue
from xrd_simulator._pickleable_object import PickleableObject
import copy
import sys

class Polycrystal(PickleableObject):

    """Represents a multi-phase polycrystal as a tetrahedral mesh where each element can be a single crystal

    Args:
        mesh (:obj:`xrd_simulator.mesh.TetraMesh`): Object representing a tetrahedral mesh which defines the 
            geometry of the sample. At instantiation it is assumed that the sample and lab coordinate systems
            are aligned.
        ephase (:obj:`numpy array`): Index of phase that elements belong to such that phases[ephase[i]] gives the
            xrd_simulator.phase.Phase object of element number i.
        eU (:obj:`numpy array`): Per element U (orinetation) matrices, (``shape=(N,3,3)``). At instantiation it
            is assumed that the sample and lab coordinate systems are aligned.
        eB (:obj:`numpy array`): Per element B (hkl to crystal mapper) matrices, (``shape=(N,3,3)``).
        phases (:obj:`list` of :obj:`xrd_simulator.phase.Phase`): List of all unique phases present in the polycrystal.

    Attributes:
        mesh_lab (:obj:`xrd_simulator.mesh.TetraMesh`): Object representing a tetrahedral mesh which defines the 
            geometry of the sample in a fixed lab frame coordinate system.
        mesh_sample (:obj:`xrd_simulator.mesh.TetraMesh`): Object representing a tetrahedral mesh which defines the 
            geometry of the sample in a sample coordinate system.
        ephase (:obj:`numpy array`): Index of phase that elements belong to such that phases[ephase[i]] gives the
            xrd_simulator.phase.Phase object of element number i.
        eU_lab (:obj:`numpy array`): Per element U (orientation) matrices in a fixed lab frame coordinate system, (``shape=(N,3,3)``).
        eU_sample (:obj:`numpy array`): Per element U (orientation) matrices in a sample coordinate system., (``shape=(N,3,3)``).
        eB (:obj:`numpy array`): Per element B (hkl to crystal mapper) matrices, (``shape=(N,3,3)``).
        phases (:obj:`list` of :obj:`xrd_simulator.phase.Phase`): List of all unique phases present in the polycrystal.

    """ 

    def __init__(self, mesh, ephase, eU, eB, phases ):

        assert eU.shape[0]==mesh.enod.shape[0], "Every crystal element must have an orientation."
        assert eB.shape[0]==mesh.enod.shape[0], "Every crystal element must have a deformation state."
        for i in range(1,3):
            assert eU.shape[i]==3
            assert eB.shape[i]==3

        # TODO: Allow specifying strain rather than eB as eB is much more 
        # non intuitive quantity.

        # Assuming sample and lab frames to be aligned at instantiation.
        self.mesh_lab = copy.deepcopy(mesh)
        self.eU_lab = eU.copy()
        self.mesh_sample = copy.deepcopy(mesh)
        self.eU_sample = eU.copy()

        self.phases = phases
        self.ephase = ephase
        self.eB     = eB

    def diffract(self, beam, detector, rigid_body_motion, min_bragg_angle=0, max_bragg_angle=None, verbose=True):
        """Compute diffraction from the rotating and translating polycrystal sample while illuminated by an xray beam. 

        The xray beam interacts with the polycrystal producing scattering regions which are stored in a detector frame.
        The scattering regions may be rendered as pixelated patterns on the detector by using a detector rendering
        option.

        Args:
            beam (:obj:`xrd_simulator.beam.Beam`): Object representing a monochromatic beam of X-rays.
            detector (:obj:`xrd_simulator.detector.Detector`): Object representing a flat rectangular detector.
            rigid_body_motion (:obj:`xrd_simulator.motion.RigidBodyMotion`): Rigid body motion object describing the
                polycrystal transformation as a funciton of time on the domain time=[0,1].
            min_bragg_angle (:obj:`float`): Minimum Bragg angle (radians) below wich to not compute diffraction. Defaults to 0.
            max_bragg_angle (:obj:`float`): Minimum Bragg angle (radians) after wich to not compute diffraction. By default the
                max_bragg_angle is approximated by wrapping the detector corners in a cone with apex at the sample centroid.
            verbose (:obj:`bool`): Prints progress. Defaults to True.

        """

        min_bragg_angle, max_bragg_angle = self._get_bragg_angle_bounds(detector, beam, min_bragg_angle, max_bragg_angle)

        for phase in self.phases:
            phase.setup_diffracting_planes(beam.wavelength, min_bragg_angle, max_bragg_angle) 

        c_0_factor   = -beam.wave_vector.dot( rigid_body_motion.rotator.K2 )
        c_1_factor   =  beam.wave_vector.dot( rigid_body_motion.rotator.K  )
        c_2_factor   =  beam.wave_vector.dot( rigid_body_motion.rotator.I + rigid_body_motion.rotator.K2 )

        scatterers = []

        proximity_intervals = beam.get_proximity_intervals(self.mesh_lab.espherecentroids, self.mesh_lab.eradius, rigid_body_motion) 

        for ei in range( self.mesh_lab.number_of_elements ):

            if verbose:
                progress_bar_message = "Computing scattering from a total of "+str(self.mesh_lab.number_of_elements)+" elements"
                progress_fraction    = float(ei+1)/self.mesh_lab.number_of_elements
                utils.print_progress(progress_fraction, message=progress_bar_message)

            # skipp elements not illuminated
            if proximity_intervals[ei][0] is None: 
                continue

            element_vertices_0 = self.mesh_lab.coord[self.mesh_lab.enod[ei],:] 
            G_0 = laue.get_G(self.eU_lab[ei], self.eB[ei], self.phases[ self.ephase[ei] ].miller_indices.T )

            c_0s  = c_0_factor.dot(G_0)
            c_1s  = c_1_factor.dot(G_0)
            c_2s  = c_2_factor.dot(G_0) + np.sum((G_0*G_0),axis=0)/2.

            for hkl_indx in range(G_0.shape[1]):
                for time in laue.find_solutions_to_tangens_half_angle_equation( c_0s[hkl_indx], c_1s[hkl_indx], c_2s[hkl_indx], rigid_body_motion.rotation_angle ):
                    if time is not None:
                        if utils.contained_by_intervals( time, proximity_intervals[ei] ):
                            element_vertices = rigid_body_motion( element_vertices_0.T, time ).T

                            #TODO: Consider plane equations representation of elements avoiding ConvexHull calls in beam.intersect()
                            scattering_region = beam.intersect( element_vertices )

                            if scattering_region is not None:
                                G = rigid_body_motion.rotate( G_0[:,hkl_indx], time )
                                scattered_wave_vector = G + beam.wave_vector                                
                                scatterer = Scatterer(  scattering_region, 
                                                        scattered_wave_vector,
                                                        beam.wave_vector,
                                                        beam.wavelength,
                                                        beam.polarization_vector,
                                                        rigid_body_motion.rotation_axis,
                                                        time, 
                                                        self.phases[ self.ephase[ei] ], 
                                                        hkl_indx )
                                scatterers.append( scatterer )
        detector.frames.append( scatterers )

    def _get_bragg_angle_bounds(self, detector, beam, min_bragg_angle, max_bragg_angle):
        """Compute a maximum Bragg angle cutof based on the beam sample interection region centroid and detector corners.

        If the beam graces or misses the sample, the sample centroid is used.
        """
        if max_bragg_angle is None:
            mesh_nodes_contained_by_beam = np.array( [c for c in self.mesh_lab.coord if beam.contains(c)] )
            if len(mesh_nodes_contained_by_beam)!=0:
                source_point = np.mean(np.array( [c for c in self.mesh_lab.coord if beam.contains(c)]), axis=0)
            else:
                source_point = self.mesh_lab.centroid
            max_bragg_angle = detector.get_wrapping_cone( beam.wave_vector, source_point )
        assert min_bragg_angle>=0, "min_bragg_angle must be greater or equal than zero"
        assert max_bragg_angle>min_bragg_angle, "max_bragg_angle must be greater than min_bragg_angle"
        return min_bragg_angle, max_bragg_angle

    def transform(self, rigid_body_motion, time):
        """Transform the polycrystal by performing a rigid body motion (translation + rotation)

        This function will update the polycrystal mesh (update in lab frame) with any dependent quantities, 
        such as face normals etc. Likewise, it will update the per element crystallite orientation 
        matrices (U) such that the updated matrix will

        Args:
            rigid_body_motion (:obj:`xrd_simulator.motion.RigidBodyMotion`): Rigid body motion object describing the
                polycrystal transformation as a funciton of time on the domain time=[0,1].
            time (:obj:`float`): Time between [0,1] at which to call the rigid body motion.

        
        """
        new_nodal_coordinates = rigid_body_motion( self.mesh_lab.coord.T, time=time ).T
        self.mesh_lab.update( new_nodal_coordinates )
        for ei in range(self.mesh_lab.number_of_elements):
            self.eU_lab[ei] = rigid_body_motion.rotate( self.eU_lab[ei], time=time )



import numpy as np 
import matplotlib.pyplot as plt
from scipy.spatial import ConvexHull, HalfspaceIntersection
from scipy.optimize import linprog
from xrd_simulator import utils, laue
from scipy.optimize import root_scalar
from xrd_simulator._pickleable_object import PickleableObject

class Beam(PickleableObject):
    """Represents a monochromatic X-ray beam as a convex polyhedra.
    
    Args:
        beam_vertices (:obj:`numpy array`): Xray-beam vertices ``shape=(N,3)``.
        xray_propagation_direction (:obj:`numpy array`): Propagation direction of X-rays ``shape=(3,)``.
        wavelength (:obj:`float`): Xray wavelength in units of angstrom.
        polarization_vector (:obj:`numpy array`): Beam linear polarization unit vector ``shape=(3,)``.
            Must be orthogonal to the xray propagation direction.

    Attributes:
        vertices (:obj:`numpy array`): Xray-beam vertices ``shape=(N,3)``.
        wavelength (:obj:`float`): Xray wavelength in units of angstrom.
        wave_vector (:obj:`numpy array`): Beam wavevector ``shape=(3,)``
        centroid (:obj:`numpy array`): Beam centroid ``shape=(3,)``
        halfspaces (:obj:`numpy array`): Beam halfspace equation coefficients ``shape=(N,3)``. 
            A point x is on the interior of the halfsplace if: halfspaces[i,:-1].dot(x) +  halfspaces[i,-1] <= 0.
        polarization_vector (:obj:`numpy array`): Beam linear polarization unit vector ``shape=(3,)``.
            Must be orthogonal to the xray propagation direction.

    """

    def __init__(self, beam_vertices, xray_propagation_direction, wavelength, polarization_vector):
        self.wave_vector = ( 2* np.pi / wavelength ) * xray_propagation_direction/np.linalg.norm( xray_propagation_direction )
        self.wavelength = wavelength
        self.set_beam_vertices(beam_vertices)
        self.polarization_vector = polarization_vector / np.linalg.norm( polarization_vector )
        assert np.allclose( np.dot( self.polarization_vector, self.wave_vector ), 0 ), "The x-ray polarisation vector is not orthogonal to the wavevector."

    def set_beam_vertices(self, beam_vertices):
        """Set the beam vertices defining the beam convex hull and update all dependent quantities.

        Args:
            beam_vertices (:obj:`numpy array`): Xray-beam vertices ``shape=(N,3)``.

        """
        #TODO: Add unit test for this function.
        ch = ConvexHull( beam_vertices )
        assert ch.points.shape[0]==ch.vertices.shape[0], "The provided beam vertices does not form a convex hull"
        self.vertices   = beam_vertices.copy()
        self.centroid   = np.mean(self.vertices, axis=0)
        self.halfspaces = ConvexHull( self.vertices ).equations

    def find_feasible_point(self, halfspaces):
        """Find a point which is clearly inside a set of halfspaces (A * point + b < 0).

        Args:
            halfspaces (:obj:`numpy array`): Halfspace equations, each row holds coefficients of a halfspace (``shape=(N,4)``).

        Returns:
            (:obj:`None`) if no point is found else (:obj:`numpy array`) point.

        """
        #NOTE: from profiling: this method is slow, beware of using it uneccesarrily.
        norm_vector = np.reshape(np.linalg.norm(halfspaces[:, :-1], axis=1), (halfspaces.shape[0], 1))
        c = np.zeros((halfspaces.shape[1],))
        c[-1] = -1
        A = np.hstack((halfspaces[:, :-1], norm_vector))
        b = - halfspaces[:, -1:]
        res = linprog(c, A_ub=A, b_ub=b, bounds=(None, None), method='highs-ipm', options={"maxiter":2000})
        if res.success and res.x[-1]>0:
            return res.x[:-1]
        else:
            return None

    def intersect( self, vertices ):
        """Compute the beam intersection with a convex polyhedra, returns a list of HalfspaceIntersections.

        Args:
            vertices (:obj:`numpy array`): Vertices of a convex polyhedra with ``shape=(N,3)``.

        Returns:
            A scipy.spatial.ConvexHull object formed from the vertices of the intersection between beam vertices and
            input vertices.

        """

        vertices_contained_by_beam = [self.contains(vertex) for vertex in vertices]
        if np.all( vertices_contained_by_beam ):
            return ConvexHull( vertices ) # The beam completely contains the input convex polyhedra

        poly_halfspace = ConvexHull( vertices ).equations
        combined_halfspaces = np.vstack( (poly_halfspace, self.halfspaces) )

        # Since find_feasible_point() is expensive it is worth checking for if the polyhedra
        # centroid is contained by the beam, being a potential cheaply computed interior_point.
        centroid = np.mean( vertices, axis=0 )
        if self.contains( centroid ):
            interior_point = centroid
        else:
            interior_point = self.find_feasible_point(combined_halfspaces)

        if interior_point is not None:
            hs = HalfspaceIntersection( combined_halfspaces , interior_point )
            return ConvexHull( hs.intersections )
        else:
            return None

    def contains( self, point ):
        """ Check if the beam contains a point.

        Args:
            point (:obj:`numpy array`): Point to evaluate ``shape=(3,)``.

        Returns:
            Boolean True if the beam contains the point.

        """
        return np.all( self.halfspaces[:,0:3].dot( point ) + self.halfspaces[:,3] < 0 )

    def get_proximity_intervals(self, sphere_centres, sphere_radius, rigid_body_motion):
        """Compute the parametric intervals t=[[t_1,t_2],[t_3,t_4],..] in which spheres are interesecting beam.

        This method can be used as a pre-checker before running the `intersect()` method on a polyhedral
        set. This avoids wasting compute resources on polyhedra which clearly do not intersect the beam.

        Args:
            sphere_centres (:obj:`numpy array`): Centroids of a spheres ``shape=(3,n)``.
            sphere_radius (:obj:`numpy array`): Radius of a spheres ``shape=(n,)``.
            rigid_body_motion (:obj:`xrd_simulator.motion.RigidBodyMotion`): Rigid body motion object describing the
                polycrystal transformation as a funciton of time on the domain time=[0,1].

        Returns:
            (:obj:`list` of :obj:`list` of :obj:`list`): Parametric ranges in which the spheres
            has an intersection with the beam. [(:obj:`None`)] if no intersection exist in ```time=[0,1]```.
            the entry at ```[i][j]``` is a list with two floats ```[t_1,t_2]``` and gives the j:th 
            intersection interval of sphere number i with the beam.

        """ 

        beam_halfplane_normals =  self.halfspaces[:,0:3]
        beam_halfplane_offsets =  self.halfspaces[:,3]

        # Precomputable factors for root equations independent of sphere data.
        a0 = rigid_body_motion.rotator.K.dot( beam_halfplane_normals.T )
        a1 = rigid_body_motion.rotator.K2.dot( beam_halfplane_normals.T )
        a2 = -beam_halfplane_normals.dot( rigid_body_motion.translation )

        all_intersections = [] # Here we will store all intersections for all spheres.

        for i in range(sphere_centres.shape[0]):

            merged_intersections = [[0., 1.]] # Here we will store all intersections for the current sphere.

            for p in range( self.halfspaces.shape[0] ):
                
                new_intersections = [] # Intersections for the sphere attached to halfplane number p.

                # Define the function we seek to have intervals in s fulfilling : intersection_function(s) < 0
                q_0  =  sphere_centres[i].dot( a0[:,p] )
                q_1  = -sphere_centres[i].dot( a1[:,p] )
                q_2  =  a2[p]
                q_3  =  sphere_centres[i].dot( beam_halfplane_normals[p,:] ) - q_1 - sphere_radius[i] + beam_halfplane_offsets[p]
                def intersection_function(time): 
                    return q_0 * np.sin( time*rigid_body_motion.rotation_angle ) + q_1 * np.cos(time*rigid_body_motion.rotation_angle) + time*q_2 + q_3

                # Find roots of intersection_function(s) by first finding its extreemal points
                brackets = self._find_brackets_of_roots(q_0, q_1, q_2, rigid_body_motion.rotation_angle, intersection_function)
                
                # Find roots numerically on the intervals 
                roots = [ root_scalar(intersection_function, method='bisect', bracket=bracket, maxiter=50).root for bracket in brackets ]
                roots.extend([0.,1.])
                interval_ends = np.sort( np.array( roots ) )
                fvals = intersection_function( (interval_ends[0:-1] + interval_ends[1:] ) /2.)

                if np.sum( fvals > 0 ) == len(fvals): 
                    merged_intersections = [None] # Always on the exterior of the beam halfplane
                    break

                if np.sum( fvals < 0 ) == len(fvals): 
                    new_intersections.append( [0.,1.] )
                    continue # Always on the interior of the beam halfplane

                # Add in all intervals where intersection_function(s) < 0
                for k in range(len(interval_ends)-1):
                    if fvals[k]<0:
                       new_intersections.append([interval_ends[k], interval_ends[k+1]])
                    else:
                        pass

                # Clean up the interval set, the sphere must intersect all halfplanes for a shared interval in s.
                merged_intersections = self._merge_intersections( merged_intersections, new_intersections )

                # The sphere will never intersect if no shared intervals are found.
                if len(merged_intersections)==0:
                    break

            all_intersections.append( merged_intersections )

        return all_intersections

    def _find_brackets_of_roots(self, q_0, q_1, q_2, rotation_angle, function):
        """Find all sub domains on time=[0,1] which are guaranteed to hold a root of function.
        """
        c_0 =  rotation_angle * q_0
        c_1 = -rotation_angle * q_1
        c_2 = q_2
        time_1, time_2 = laue.find_solutions_to_tangens_half_angle_equation( c_0, c_1, c_2, rotation_angle )
        search_intervals = np.array( [s for s in [0, time_1, time_2, 1] if s is not None] )
        f = function( search_intervals )
        indx = 0
        brackets = []
        for i in range(len(search_intervals)-1):
            if np.sign( f[indx] )!=np.sign( f[i+1] ):
                brackets.append( [search_intervals[indx],search_intervals[i+1]] )
                indx = i+1
        return brackets

    def _merge_intersections(self, merged_intersections, new_intersections):
        """Return intersection of merged_intersections and new_intersections

        NOTE: Assumes no overlaps inside the brackets of merged_intersections
        and new_intersections. I.e they are minimal descriptions of their
        respective point sets.
        """
        new_intervals = []
        for mi in merged_intersections:
            for ni in new_intersections:
                intersection = self._intersection(mi, ni)
                if intersection is not None:
                    new_intervals.append( intersection )
        return new_intervals

    def _intersection(self, bracket1, bracket2):
        """Find the intersection between two simple bracket intervals of form [start1,end1] and [start2,end2].
        """
        if (bracket2[0] <= bracket1[0] <= bracket2[1]) or (bracket2[0] <= bracket1[1] <= bracket2[1]):
            points = np.sort([ bracket2[0],bracket1[0],bracket2[1],bracket1[1] ])
            return [points[1],points[2]]
        else:
            return None


            



import unittest
import numpy as np
from xrd_simulator.detector import Detector
from xrd_simulator.phase import Phase
from xrd_simulator.scatterer import Scatterer
from scipy.spatial import ConvexHull
from xrd_simulator.beam import Beam
import os
import matplotlib.pyplot as plt

class TestDetector(unittest.TestCase):

    def setUp(self):
        self.pixel_size_z = 50.
        self.pixel_size_y = 40.
        self.detector_size = 10000.
        self.d0 = np.array([1,0,0])*self.detector_size
        self.d1 = np.array([1,1,0])*self.detector_size
        self.d2 = np.array([1,0,1])*self.detector_size
        self.detector = Detector( self.pixel_size_z, self.pixel_size_y, self.d0, self.d1, self.d2 )

    def test_init(self):

        for o,otrue in zip(self.detector.d0, np.array([1,0,0])*self.detector_size):
            self.assertAlmostEqual(o,otrue ,msg="detector origin is incorrect")

        for z,ztrue in zip(self.detector.zdhat, np.array([0,0,1])):
            self.assertAlmostEqual(z, ztrue ,msg="zdhat is incorrect")

        for y,ytrue in zip(self.detector.ydhat, np.array([0,1,0])):
            self.assertAlmostEqual(y, ytrue ,msg="ydhat is incorrect")

        self.assertAlmostEqual(self.detector.zmax, self.detector_size, msg="Bad detector dimensions in zmax")
        self.assertAlmostEqual(self.detector.ymax, self.detector_size, msg="Bad detector dimensions in ymax")

        for n,ntrue in zip(self.detector.normal, np.array([-1,0,0])):
            self.assertAlmostEqual(n, ntrue ,msg="Bad detector normal")

    def test_centroid_render(self):
        v = self.detector.ydhat + self.detector.zdhat
        v = v/np.linalg.norm(v)
        verts1 = np.array([[0,0,0],
                          [0,0,1],
                          [0,1,0],
                          [1,0,0]]) + v*np.sqrt(2)*self.detector_size/2. # tetra at detector centre
        ch1 = ConvexHull( verts1 )
        verts2 = np.array([[0,0,0],
                          [0,0,1],
                          [0,1,0],
                          [1,0,0]]) + 2*v*np.sqrt(2)*self.detector_size # tetra out of detector bounds
        ch2 = ConvexHull( verts2 )
        wavelength = 1.0

        incident_wave_vector = 2*np.pi* np.array([1,0,0])/(wavelength)
        scattered_wave_vector = self.d0 + self.pixel_size_y*3*self.detector.ydhat + self.pixel_size_z*2*self.detector.zdhat
        scattered_wave_vector = 2*np.pi*scattered_wave_vector/(np.linalg.norm(scattered_wave_vector)*wavelength)

        data = os.path.join( os.path.join(os.path.dirname(__file__), 'data' ), 'Fe_mp-150_conventional_standard.cif' )
        unit_cell = [3.64570000, 3.64570000, 3.64570000, 90.0, 90.0, 90.0]
        sgname = 'Fm-3m' # Iron
        phase   = Phase(unit_cell, sgname, path_to_cif_file=data)
        phase.setup_diffracting_planes(wavelength, 0, 20*np.pi/180)

        scatterer1 = Scatterer( ch1, 
                                scattered_wave_vector=scattered_wave_vector, 
                                incident_wave_vector=incident_wave_vector, 
                                wavelength=wavelength,
                                incident_polarization_vector=np.array([0,1,0]), 
                                rotation_axis=np.array([0,0,1]),
                                time=0, 
                                phase=phase, 
                                hkl_indx=0 )
        scatterer2 = Scatterer( ch2, 
                                scattered_wave_vector=scattered_wave_vector, 
                                incident_wave_vector=incident_wave_vector, 
                                wavelength=wavelength,
                                incident_polarization_vector=np.array([0,1,0]), 
                                rotation_axis=np.array([0,0,1]),
                                time=0, 
                                phase=phase, 
                                hkl_indx=0 )

        self.detector.frames.append([scatterer1, scatterer2])
        diffraction_pattern = self.detector.render(frame_number=0, lorentz=False, polarization=False, structure_factor=False, method="centroid")
        
        # the sample sits at the centre of the detector.
        expected_z_pixel = int( self.detector_size / (2*self.pixel_size_z) ) + 2
        expected_y_pixel = int( self.detector_size / (2*self.pixel_size_y) ) + 3

        self.assertAlmostEqual(diffraction_pattern[expected_z_pixel,expected_y_pixel], ch1.volume, msg="detector rendering did not capture scatterer")
        self.assertAlmostEqual(np.sum(diffraction_pattern), ch1.volume, msg="detector rendering captured out of bounds scatterer")

        # Try rendering with advanced intensity model
        diffraction_pattern = self.detector.render(frame_number=0, lorentz=True, polarization=False, structure_factor=False)
        self.assertTrue(diffraction_pattern[expected_z_pixel,expected_y_pixel]!=ch1.volume, msg="detector rendering did not use lorentz factor")
        
        diffraction_pattern = self.detector.render(frame_number=0, lorentz=False, polarization=True, structure_factor=False)
        self.assertTrue(diffraction_pattern[expected_z_pixel,expected_y_pixel]!=ch1.volume, msg="detector rendering did not use polarization factor")
        
        diffraction_pattern = self.detector.render(frame_number=0, lorentz=False, polarization=False, structure_factor=True)
        self.assertTrue(diffraction_pattern[expected_z_pixel,expected_y_pixel]!=ch1.volume, msg="detector rendering did not use structure_factor factor")

    def test_projection_render(self):
        
        # Convex hull of a sphere placed at the centre of the detector
        phi, theta  = np.meshgrid( np.linspace( 0, 2*np.pi, 25 ), np.linspace( 0, 2*np.pi, 25 ), indexing='ij' )
        r     = 1.0*self.detector_size/4.
        x,y,z = r*np.cos( phi )*np.sin( theta ),r*np.sin( phi )*np.sin( theta ),r*np.cos( theta )
        hull_points = np.array( [x.flatten(),y.flatten(),z.flatten()] ).T
        v = self.detector.ydhat + self.detector.zdhat
        v = v/np.linalg.norm(v)
        sphere_hull = ConvexHull( hull_points + v*np.sqrt(2)*self.detector_size/2. )

        # The spherical scatterer forward scatters.
        wavelength = 1.0
        incident_wave_vector = 2*np.pi* np.array([1,0,0])/wavelength
        scattered_wave_vector = 2*np.pi* np.array([1,0,0])/wavelength
        scattered_wave_vector = 2*np.pi*scattered_wave_vector/(np.linalg.norm(scattered_wave_vector)*wavelength)

        # The spherical scatterer is composed of Fe_mp-150 (a pure iron crystal)
        data = os.path.join( os.path.join(os.path.dirname(__file__), 'data' ), 'Fe_mp-150_conventional_standard.cif' )
        unit_cell = [3.64570000, 3.64570000, 3.64570000, 90.0, 90.0, 90.0]
        sgname = 'Fm-3m' # Iron
        phase   = Phase(unit_cell, sgname, path_to_cif_file=data)
        phase.setup_diffracting_planes(wavelength, 0, 20*np.pi/180)

        scatterer = Scatterer(  sphere_hull, 
                                scattered_wave_vector=scattered_wave_vector, 
                                incident_wave_vector=incident_wave_vector, 
                                wavelength=wavelength,
                                incident_polarization_vector=np.array([0,1,0]), 
                                rotation_axis=np.array([0,0,1]),
                                time=0, 
                                phase=phase,
                                hkl_indx=0 )

        self.detector.frames.append([scatterer])
        diffraction_pattern = self.detector.render(frame_number=0, lorentz=False, polarization=False, structure_factor=False, method="project")

        projected_summed_intensity = np.sum(diffraction_pattern)
        relative_error             = np.abs( scatterer.volume - projected_summed_intensity ) / scatterer.volume
        self.assertLessEqual( relative_error, 1e-4, msg="Projected mass does not match the hull volume" )

        index = np.where(diffraction_pattern==np.max(diffraction_pattern))
        self.assertEqual( index[0][0], self.detector_size//(2*self.pixel_size_z), msg="Projected mass does not match the hull volume" )
        self.assertEqual( index[1][0], self.detector_size//(2*self.pixel_size_y), msg="Projected mass does not match the hull volume" )

        no_pixels_z = int( self.detector_size//self.pixel_size_z )
        no_pixels_y = int( self.detector_size//self.pixel_size_y )


        for i in range(no_pixels_z):
            for j in range(no_pixels_y):
                zd = i*self.pixel_size_z
                yd = j*self.pixel_size_y
                if (zd - self.detector_size/2.)**2 + (yd - self.detector_size/2.)**2 > (1.01*r)**2:
                    self.assertAlmostEqual( diffraction_pattern[i,j], 0 )
                elif (zd - self.detector_size/2.)**2 + (yd - self.detector_size/2.)**2 < (0.99*r)**2:
                    self.assertGreater( diffraction_pattern[i,j], 0 )

    def test_get_intersection(self):

        # central normal algined ray
        ray_direction = np.array([2.23,0.,0.])
        source_point  = np.array([0.,0.,0.])
        z, y  = self.detector.get_intersection(ray_direction, source_point)
        self.assertAlmostEqual(z,0,msg="central detector-normal algined ray does not intersect at 0")
        self.assertAlmostEqual(y,0,msg="central detector-normal algined ray does not intersect at 0")

        # translate the ray
        source_point  += self.detector.ydhat * self.pixel_size_y
        source_point  -= self.detector.zdhat * 2 * self.pixel_size_z
        z, y  = self.detector.get_intersection(ray_direction, source_point)
        self.assertAlmostEqual(z, -2*self.pixel_size_z, msg="translated detector-normal algined ray does not intersect properly")
        self.assertAlmostEqual(y, self.pixel_size_y, msg="translated detector-normal algined ray does not intersect properly")

        # tilt the ray
        ang = np.arctan(self.pixel_size_y/self.detector_size)
        frac = np.tan(ang)*np.linalg.norm(ray_direction)
        ray_direction += self.detector.ydhat * frac * 3
        z, y  = self.detector.get_intersection(ray_direction, source_point)
        self.assertAlmostEqual(z, -2*self.pixel_size_z, msg="translated and tilted ray does not intersect properly")
        self.assertAlmostEqual(y, 4*self.pixel_size_y, msg="translated and tilted ray does not intersect properly")

    def test_contains(self):
        c1 = self.detector.contains(self.detector_size/10., self.detector_size/5.)
        self.assertTrue(c1, msg="detector does no contain included point")
        c2= self.detector.contains(-self.detector_size/8., self.detector_size/3.)
        self.assertTrue(not c2, msg="detector contain negative points")
        c4= self.detector.contains(self.detector_size*2*self.pixel_size_z, self.detector_size/374.)
        self.assertTrue(not c4, msg="detector contain out of bounds points")

    def test_get_wrapping_cone(self):
        wavelength = 1.0
        k = 2 * np.pi * np.array([1,0,0]) / wavelength
        source_point  = (self.detector.zdhat + self.detector.ydhat) * self.detector_size/2.
        opening_angle = self.detector.get_wrapping_cone(k, source_point)

        normed_det_center = (source_point + self.d0)/np.linalg.norm(source_point + self.d0)
        normed_det_origin = self.d0/np.linalg.norm(self.d0)
        expected_angle = np.arccos( np.dot( normed_det_center, normed_det_origin ) ) / 2.

        self.assertAlmostEqual(opening_angle, expected_angle, msg="detector centered wrapping cone has faulty opening angle")

        source_point  = (self.detector.zdhat + self.detector.ydhat) * self.detector_size/2.
        source_point -= self.detector.zdhat*10*self.pixel_size_z
        source_point -= self.detector.ydhat*10*self.pixel_size_y
        opening_angle = self.detector.get_wrapping_cone(k, source_point)
        self.assertGreaterEqual(opening_angle, expected_angle, msg="detector off centered wrapping cone has opening angle")

if __name__ == '__main__':
    unittest.main()


"""Run all unittests in the tests folder"""
from unittest import TestLoader, TestResult
from pathlib import Path
from xrd_simulator.utils import _HiddenPrints

def run_tests():
    test_loader = TestLoader()
    test_result = TestResult()

    test_directory = str(Path(__file__).resolve().parent)

    test_suite = test_loader.discover(test_directory, pattern='test_*.py')

    test_suite.run(result=test_result)

    print("\n")
    if test_result.wasSuccessful():
        print(" All unittests ran successfully !")
        exit(0)
    else:
        print(" Unittests Failed !")
        print(test_result.failures)
        print(test_result.errors)
        exit(-1)

if __name__ == '__main__':
    run_tests() # Runs all unittests


import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse import csr_matrix
from scipy.spatial import ConvexHull
from numba import njit
import pygalmesh
import meshio
import miniball

class TetraMesh(object): #TODO: add unit tests
    """Defines a 3D tetrahedral finite element type basis by subclassing :obj:`Basis`. 

    Attributes:
        coord (:obj:`numpy array`): Nodal coordinates, shape=(nenodes, 3). Each row in coord defines the 
            coordinates of a mesh node.
        enod (:obj:`numpy array`): Tetra element nodes shape=(nelm, nenodes).e.g enod[i,:] gives
            the nodal indices of element i.
        dof (:obj:`numpy array`): Per node degrees of freedom, i.e dof[i,:] 
            gives the degrees of freedom of node i.
        efaces (:obj:`numpy array`): Element faces nodal indices, shape=(nelm, nenodes, 3).
            e.g efaces[i,j,:] gives the nodal indices of face j of element i.
        enormals (:obj:`numpy array`): Element faces outwards normals (nelm, nefaces, 3).
            e.g enormals[i,j,:] gives the normal of face j of element i.
        ecentroids (:obj:`numpy array`): Per element centroids, shape=(nelm, 3).
        eradius (:obj:`numpy array`): Per element bounding radius, shape=(nelm, 1).
        ecmat (:obj:`numpy array`): Per element interpolation matrix, shape=(nelm, 4, 4). When
            multiplied on a coordinate array, :obj:`np.array([1,x,y,z])`, the interpolated value 
            at x,y,z is found, given that x,y,z is contained by the corresponding element. 

    """

    def __init__(self):
        self._mesh       = None
        self.coord       = None
        self.enod        = None
        self.dof         = None
        self.efaces      = None
        self.enormals    = None
        self.ecentroids  = None
        self.eradius     = None
        self.espherecentroids = None
        self.evolumes    = None
        self.ecmat       = None
        self.centroid    = None
        self.number_of_elements = None

    @classmethod
    def _build_tetramesh(cls, mesh):
        tetmesh = cls()
        tetmesh._mesh = mesh
        tetmesh._set_fem_matrices()
        tetmesh._expand_mesh_data()
        return tetmesh

    @classmethod
    def load(cls, path):
        """Load a mesh from a saved mesh file set using `the meshio package`_:

        .. _the meshio package: https://github.com/nschloe/meshio

        Args:
            file (:obj:`str`): Absolute path to save the mesh at (without .xdmf extension)

        """
        mesh = meshio.read( path )
        return cls._build_tetramesh( mesh)

    @classmethod
    def generate_mesh_from_vertices(cls, coord, enod):
        """Generate a mesh from vertices using `the meshio package`_:

        .. _the meshio package: https://github.com/nschloe/meshio

        Args:
            coord (:obj:`numpy array`): Nodal coordinates, shape=(nenodes, 3). Each row in coord defines the 
                coordinates of a mesh node.
            enod (:obj:`numpy array`): Tetra element nodes shape=(nelm, nenodes).e.g enod[i,:] gives
                the nodal indices of element i.

        """
        mesh = meshio.Mesh( coord, [("tetra", enod)] )
        return cls._build_tetramesh( mesh)

    @classmethod
    def generate_mesh_from_levelset(cls, level_set, bounding_radius, max_cell_circumradius):
        """Generate a mesh from a level set using `the pygalmesh package`_:

        .. _the pygalmesh package: https://github.com/nschloe/pygalmesh

        Args:
            level_set (:obj:`callable`): Level set, level_set(x) should give a negative output on the exterior
                of the mesh and positive on the interior.
            bounding_radius (:obj:`float`): Bounding radius of mesh.
            max_cell_circumradius (:obj:`float`): Bound for element radii.

        """

        class LevelSet(pygalmesh.DomainBase):
            def __init__(self):
                super().__init__()
                self.eval = level_set
                self.get_bounding_sphere_squared_radius = lambda : bounding_radius**2

        mesh = pygalmesh.generate_mesh( LevelSet(),
                                        max_cell_circumradius=max_cell_circumradius, 
                                        verbose=False)

        return cls._build_tetramesh( mesh)

    @classmethod
    def generate_mesh_from_numpy_array(cls, array, voxel_size, max_cell_circumradius):
        """Generate a mesh from a numpy array using `the pygalmesh package`_:

        .. _the pygalmesh package: https://github.com/nschloe/pygalmesh

        Args:
            array (:obj:`numpy array`): Numpy array to generate mesh from.
            voxel_size (:obj:`float`): Dimension of array voxels.
            max_cell_circumradius (:obj:`float`): Bound for element radii.

        """

        mesh = pygalmesh.generate_from_array( array, [voxel_size]*3,
                                            max_cell_circumradius=max_cell_circumradius, 
                                            verbose=False )
        return cls._build_tetramesh( mesh)

    def _set_fem_matrices(self):
        """Extract and set mesh FEM matrices from pygalmesh object.
        """
        self.coord      = np.array(self._mesh.points)
        self.enod       = np.array(self._mesh.cells_dict['tetra'])
        self.dof        = np.arange(0,self.coord.shape[0]*3).reshape(self.coord.shape[0],3)
        self.coefficients = np.zeros(self.coord.shape)
        self.number_of_elements = self.enod.shape[0]

    def _expand_mesh_data(self):
        """Compute extended mesh quantities such as element faces and normals.
        """
        self.efaces          = self._compute_mesh_faces( self.enod )
        self.enormals        = self._compute_mesh_normals( self.coord, self.enod, self.efaces )
        self.ecentroids      = self._compute_mesh_centroids( self.coord, self.enod ) 
        self.eradius,self.espherecentroids = self._compute_mesh_spheres( self.coord, self.enod )
        self.ecmat           = self._compute_mesh_interpolation_matrices( self.enod, self.coord )
        self.centroid        = np.mean(self.ecentroids, axis=0)
        self.evolumes        = self._compute_mesh_volumes( self.enod, self.coord )

        #TODO: considering leveraging this in beam.py for speed
        # self.econvexhulls    = [ ConvexHull( self.coord[nodes,:] ) for nodes in self.enod ] 

    def update( self, new_nodal_coordinates ):
        """Update the mesh coordinates and any dependent quantities by changing the node coordinates.
        """
        self._mesh.points = new_nodal_coordinates
        self._set_fem_matrices()
        self._expand_mesh_data()

    def get_bounding_sphere_radius(self, node):
        """This method overrides :meth:`Basis.get_bounding_sphere_radius`.
        """
        elements = np.where( self.enod==node )[0]
        nodes = np.unique( self.enod[elements,:] )
        return np.max( np.linalg.norm( self.coord[nodes,:] - self.coord[node,:], axis=1 ) ) + 1e-8

    def _compute_mesh_interpolation_matrices( self, enod, coord ):
        """compute tetra element inverse C matrices.
        """
        interpolation_matrices = np.zeros((enod.shape[0],4,4))
        for element in range(enod.shape[0]):
            ec = coord[enod[element],:]
            V = np.ones( (ec.shape[0],ec.shape[0]) )
            V[1:,:] = ec.T[:,:]
            interpolation_matrices[element,:,:] = np.linalg.inv( V ).T
        return interpolation_matrices

    def _compute_interpolation_matrices( self, enod, coord ):
        """compute tetra element inverse C matrices.
        """
        interpolation_matrices = np.zeros((enod.shape[0],4,4))
        for element in range(enod.shape[0]):
            ec = coord[enod[element],:]
            V = np.ones( (ec.shape[0],ec.shape[0]) )
            V[1:,:] = ec.T[:,:]
            interpolation_matrices[element,:,:] = np.linalg.inv( V ).T
        return interpolation_matrices

    def _compute_mesh_faces(self, enod):
        """Compute all element faces nodal numbers.
        """
        efaces = np.zeros( (enod.shape[0], 4, 3), dtype=int )
        for i in range( enod.shape[0] ):
            # nodal combinations defining 4 unique planes in a tet.
            permutations = [ [0,1,2], [0,1,3], [0,2,3], [1,2,3] ]
            for j,perm in enumerate( permutations ):
                efaces[i,j,:] = enod[i,perm]
        return efaces

    def _compute_mesh_normals(self, coord, enod, efaces):
        """Compute all element faces outwards unit vector normals.
        """
        enormals = np.zeros( (enod.shape[0], 4, 3) )
        for i in range( enod.shape[0] ):
            ec = coord[enod[i,:], :]
            ecentroid = np.mean( ec, axis=0 )
            for j in range( efaces.shape[1] ):
                ef = coord[efaces[i,j,:],:]
                enormals[i,j,:] = self._compute_plane_normal( ef, ecentroid )
        return enormals

    def _compute_plane_normal( self, points, centroid ):
        """Compute plane normal (outwards refering to centroid).
        """
        v1 = points[1,:] - points[0,:]
        v2 = points[2,:] - points[0,:]
        n  = np.cross(v1, v2)                            # define a vector perpendicular to the plane.
        n  = n*np.sign( n.dot(points[0,:] - centroid) )  # set vector direction outwards from centroid.
        return n/np.linalg.norm(n)                       # normalised vector and return.

    def _compute_mesh_centroids(self, coord, enod):
        """Compute centroids of elements.
        """
        ecentroids = np.zeros((enod.shape[0],3))
        for i in range( enod.shape[0] ):
            ec = coord[enod[i,:], :]
            ecentroids[i,:] = np.sum( ec, axis=0 )/ec.shape[0]
        return ecentroids

    def _compute_mesh_volumes( self, enod, coord ):
        """Compute per element enclosed volume.
        """
        evolumes = np.zeros((enod.shape[0],))
        for i in range( enod.shape[0] ):
            ec = coord[enod[i,:], :]
            a = ec[1] - ec[0]
            b = ec[2] - ec[0]
            c = ec[3] - ec[0]
            evolumes[i] = (1/6.)*np.dot( np.cross( a, b ), c)
        return evolumes

    def _compute_mesh_spheres(self, coord, enod):
        """Compute per element minimal bounding spheres.
        """
        eradius = np.zeros((enod.shape[0],))
        espherecentroids = np.zeros((enod.shape[0],3))
        for i in range( enod.shape[0] ):
            ec = coord[enod[i,:], :]
            c, r = miniball.get_bounding_ball(ec, epsilon=1e-8)
            espherecentroids[i] = c
            eradius[i] = np.sqrt( r )
            # Whats this .. no trust ...
            for c in ec:
                assert (c-espherecentroids[i]).dot(c-espherecentroids[i]) <= (eradius[i]*1.001)**2, "error = "+str( (c-espherecentroids[i]).dot(c-espherecentroids[i]) - eradius[i]**2 )
        return eradius, espherecentroids

    def __call__(self, X, Y, Z, dim='all'):
        """This method overrides :meth:`Basis.__call__`.
        """
        shape = X.shape
        xs, ys, zs = X.flatten().astype(np.float64), Y.flatten().astype(np.float64), Z.flatten().astype(np.float64)

        if dim=='all':
            values = _get_interpolation_values_nd( xs, ys, zs,
                                            self.ecentroids.astype(np.float64),
                                            self.eradius.astype(np.float64),
                                            self.enormals,
                                            self.coord.astype(np.float64),
                                            self.efaces,
                                            self.ecmat,
                                            self.enod,
                                            self.coefficients.astype(np.float64) )
            return values.reshape(shape+(3,))
        else:
            values = _get_interpolation_values_1d( xs, ys, zs,
                                            self.ecentroids.astype(np.float64),
                                            self.eradius.astype(np.float64),
                                            self.enormals,
                                            self.coord.astype(np.float64),
                                            self.efaces,
                                            self.ecmat,
                                            self.enod,
                                            self.coefficients.astype(np.float64),
                                            dim )
            return values.reshape(shape)

    def save(self, file, element_data=None):
        """Save the tetra mesh to .xdmf paraview readable format for visualization.

        Args:
            file (:obj:`str`): Absolute path to save the mesh in .xdmf format.
        """
        #TODO: add end to end test
        if element_data is not None:
            for key in element_data:
                self._mesh.cell_data[key] = np.array([element_data[key]])

        if file.endswith(".xdmf"):
            self._mesh.write(file)
        else:
            self._mesh.write(file+".xdmf")

@njit
def _get_candidate_elements( point, ecentroids, eradius ):
    """Find all elements that could contain a point based on their bounding radii.
    This is a just in time compiled helper function for :obj:`TetraMesh` used for fast interpolation.
    """
    distance_vectors = ecentroids - point
    euclidean_distances = np.sum( distance_vectors*distance_vectors, axis=1 )
    candidate_elements = np.where( euclidean_distances <= eradius**2 )[0] 
    return candidate_elements[ np.argsort( euclidean_distances[candidate_elements] ) ]   

@njit
def _is_in_element( element, point, efaces, enormals, coord ):
    """Check if a point is contained by an element.
    This is a just in time compiled helper function for :obj:`TetraMesh` used for fast interpolation.
    """
    for face in range(efaces.shape[1]):
        face_node = efaces[element, face, 0]
        if (coord[face_node,:] - point).dot( enormals[element,face,:] ) < 0:
            break
        if face==efaces.shape[1]-1:
            return True
    return False

@njit
def _find_element_owner( point, ecentroids, eradius, enormals, coord, efaces, enod, element_guess):
    """Find the element that contains a point.
    This is a just in time compiled helper function for :obj:`TetraMesh` used for  fast interpolation.
    """ 
    if _is_in_element(element_guess, point, efaces, enormals, coord):
        return element_guess

    candidate_elements  = _get_candidate_elements(point, ecentroids, eradius)
    for element in candidate_elements:
        if _is_in_element(element, point, efaces, enormals, coord):
            return element

    return -1

@njit
def _get_interpolation_values_nd( xs, ys, zs, ecentroids, eradius, enormals, coord, efaces, ecmat, enod, coefficients ):
    """Compute mesh interpolated vector values at a series of coordinates.
    This is a just in time compiled helper function for :obj:`TetraMesh` used for fast interpolation.
    """
    values = np.zeros((len(xs),3))
    element_guess = 0
    for i,(x,y,z) in enumerate(zip(xs, ys, zs)):
        element = _find_element_owner( np.array([x,y,z]), ecentroids, eradius, enormals, coord, efaces, enod, element_guess)
        if element<0: 
            values[i,:] = 0
        else:
            element_nodes = enod[element, :]
            element_coefficients = coefficients[element_nodes, :]
            values[i,:] = element_coefficients.T.dot(  ecmat[element,:,:].T.dot(np.array([1,x,y,z])) )
            element_guess = element
    return values

@njit
def _get_interpolation_values_1d( xs, ys, zs, ecentroids, eradius, enormals, coord, efaces, ecmat, enod, coefficients, dim ):
    """Compute mesh interpolated scalar values at a series of coordinates.
    This is a just in time compiled helper function for :obj:`TetraMesh` used for fast interpolation.
    """ 
    values = np.zeros((len(xs),))
    element_guess = 0
    for i,(x,y,z) in enumerate(zip(xs, ys, zs)):
        element = _find_element_owner( np.array([x,y,z]), ecentroids, eradius, enormals, coord, efaces, enod, element_guess)
        if element<0: 
            values[i] = 0
        else:
            element_nodes = enod[element, :]
            element_coefficients = coefficients[element_nodes, dim]
            values[i] = element_coefficients.T.dot(  ecmat[element,:,:].T.dot(np.array([1,x,y,z])) )
            element_guess = element
    return values


"""General package internal utility functions.
"""
import os
import sys

import numpy as np
from numba import njit
from xfab import tools


class _HiddenPrints:
    """Simple class to enable running code without printing using python with statements.

    This is a hack to suppress printing from imported packages over which we do not have full
    controll. (xfab.tools)
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def contained_by_intervals(value, intervals):
    """Assert if a float ``value`` is contained by any of a number of ``intervals``.
    """
    for bracket in intervals:
        if value >= bracket[0] and value <= bracket[1]:
            return True
    return False


def cif_open(cif_file):
    """Helper function to be able to use the ``.CIFread`` of ``xfab``.
    """
    from CifFile import ReadCif
    cif_dict = ReadCif(cif_file)
    return cif_dict[list(cif_dict.keys())[0]]


def print_progress(progress_fraction, message):
    """Print a progress bar in the executing shell terminal.

    Args:
        progress_fraction (:obj:`float`): progress between 0 and 1.
        message (:obj:`str`): Optional message prepend the loading bar with. (max 55 characters)

    """
    assert len(message) <= 55., "Message to print is too long, max 55 characters allowed."
    progress_in_precent = np.round(100 * progress_fraction, 1)
    progress_bar_length = int(progress_fraction * 40)
    sys.stdout.write("\r{0}{1} | {2}>{3} |".format(message, " " * (55 - len(message)), "=" *
                                                   progress_bar_length, " " *
                                                   (40 - progress_bar_length)) + " " +
                     str(progress_in_precent) + "%")
    if progress_fraction != 1.0:
        sys.stdout.flush()
    else:
        sys.stdout.write("\n")


@njit
def clip_line_with_convex_polyhedron(line_points, line_direction, plane_points, plane_normals):
    """Compute lengths of parallel lines clipped by a convex polyhedron defined by 2d planes.

        For algorithm description see: Mike Cyrus and Jay Beck. “Generalized two- and three-
        dimensional clipping”. (1978) The algorithms is based on solving orthogonal equations and
        sorting the resulting plane line interestion points to find which are entry and which are
        exit points through the convex polyhedron.

        Args:
            line_points (:obj:`numpy array`): base points of rays
                (exterior to polyhedron), ``shape=(n,3)``
            line_direction  (:obj:`numpy array`): normalized ray direction
                (all rays have the same direction),  ``shape=(3,)``
            plane_points (:obj:`numpy array`): point in each polyhedron face plane, ``shape=(m,3)``
            plane_normals (:obj:`numpy array`): outwards element face normals. ``shape=(m,3)``

        Returns:
            clip_lengths (:obj:`numpy array`) : intersection lengths.  ``shape=(n,)``

    """
    clip_lengths = np.zeros((line_points.shape[0],))
    t_2 = np.dot(plane_normals, line_direction)
    te_mask = t_2 < 0
    tl_mask = t_2 > 0
    for i, line_point in enumerate(line_points):

        # find parametric line-plane intersection based on orthogonal equations
        t_1 = np.sum(np.multiply(plane_points - line_point, plane_normals), axis=1)

        # Zero division for a ray parallel to plane, numpy gives np.inf so it is ok!
        t_i = t_1 / t_2

        # Sort intersections points as potential entry and exit points
        t_e = np.max(t_i[te_mask])
        t_l = np.min(t_i[tl_mask])

        if t_l > t_e:
            clip_lengths[i] = t_l - t_e

    return clip_lengths


def alpha_to_quarternion(alpha_1, alpha_2, alpha_3):
    """Generate a unit quarternion by providing spherical angle coordinates on the S3 ball.

    Args:
        alpha_1,alpha_2,alpha_3 (:obj:`numpy array` or :obj:`float`): Radians. ``shape=(N,4)``

    Returns:
        (:obj:`numpy array`) Rotation as unit quarternion ``shape=(N,4)``

    """
    sin_alpha_1, sin_alpha_2 = np.sin(alpha_1), np.sin(alpha_2)
    return np.array([np.cos(alpha_1),
                     sin_alpha_1 * sin_alpha_2 * np.cos(alpha_3),
                     sin_alpha_1 * sin_alpha_2 * np.sin(alpha_3),
                     sin_alpha_1 * np.cos(alpha_2)]).T


def lab_strain_to_lattice_matrix(strain_tensor, crystal_orientation, unit_cell):
    """Take a strain tensor in crystal coordinates and produce the lattice matrix (B matrix).

    Args:
        strain_tensor (:obj:`numpy array`): Symmetric strain tensor in crystal
            coordinates. ``shape=(3,3)``
        crystal_orientation (:obj:`numpy array`): Unitary crystal orientation matrix.
            ``crystal_orientation`` maps from crystal to lab coordinates. ``shape=(3,3)``
        unit_cell (:obj:`list` of :obj:`float`): Crystal unit cell representation of the form
            [a,b,c,alpha,beta,gamma], where alpha,beta and gamma are in units of degrees while
            a,b and c are in units of anstrom.

    Returns:
        (:obj:`numpy array`) Lattice matrix (B matrix)``shape=(3,3)``

    """
    crystal_strain = np.dot(
        crystal_orientation.T, np.dot(
            strain_tensor, crystal_orientation))
    lattice_matrix = tools.epsilon_to_b([crystal_strain[0, 0],
                                         crystal_strain[0, 1],
                                         crystal_strain[0, 2],
                                         crystal_strain[1, 1],
                                         crystal_strain[1, 2],
                                         crystal_strain[2, 2]],
                                        unit_cell)
    return lattice_matrix


def diffractogram( diffraction_pattern, det_centre_z, det_centre_y, binsize=1.0 ):
    """Compute diffractogram from pixelated diffraction pattern.

    Args:
        diffraction_pattern (:obj:`numpy array`): Pixelated diffraction pattern``shape=(m,n)``
        det_centre_z (:obj:`numpy array`): Intersection pixel coordinate between
                 beam centroid line and detector along z-axis.
        det_centre_y (:obj:`list` of :obj:`float`): Intersection pixel coordinate between
                 beam centroid line and detector along y-axis.
        binsize  (:obj:`list` of :obj:`float`): Histogram binsize. (Detector pixels are integrated
            radially around the azimuth)

    Returns:
        (:obj:`tuple`) with ``bin_centres`` and ``histogram``.

    """
    m,n = diffraction_pattern.shape
    max_radius = np.max([m,n])
    bin_centres = np.arange( 0, int(max_radius+1), binsize )
    histogram = np.zeros( (len(bin_centres), ) )
    for i in range(m):
        for j in range(n):
            radius = np.sqrt( (i-det_centre_z)**2 + (j-det_centre_y)**2 )
            bin_index = np.argmin( np.abs( bin_centres - radius ) )
            histogram[bin_index] += diffraction_pattern[i,j]
    clip_index=len(histogram)-1
    for k in range(len(histogram)-1, 0, -1):
        if histogram[k]!=0: break
        else: clip_index = k
    clip_index = np.min([clip_index+m//10, len(histogram)-1])
    return bin_centres[0:clip_index], histogram[0:clip_index]


import numpy as np 
import matplotlib.pyplot as plt

class Scatterer(object):

    """Defines a scattering single crystal as a convex polyhedra.

    Args:
        convex_hull (:obj:`scipy.spatial.ConvexHull`): Object describing the convex hull of the self.
        kprime (:obj:`string`): Scattering vector, i.e the wavevector pointing in the direction of diffraction.
        s (:obj:`float`): Parametric value in range [0,1] where 0 corresponds to a beam with wavevector k1
            while s=1 to a beam with wavevector k2. The s value of the scatterer implies what detector
            position is applicable during diffraction.
        phase (:obj:`Phase`): The Phase object representing the material of the self.
        hkl_indx (:obj:`int`): Index of Miller index in the `phase.miller_indices` list.

    Attributes:
        convex_hull (:obj:`scipy.spatial.ConvexHull`): Object describing the convex hull of the self.
        kprime (:obj:`string`): Scattering vector, i.e the wavevector pointing in the direction of diffraction.
        s (:obj:`float`): Parametric value in range [0,1] where 0 corresponds to a beam with wavevector k1
            while s=1 to a beam with wavevector k2. The s value of the scatterer implies what detector
            position is applicable during diffraction.

    """ 

    def __init__(self, convex_hull, scattered_wave_vector, incident_wave_vector, wavelength, incident_polarization_vector, rotation_axis, time, phase, hkl_indx ):
        self.convex_hull           = convex_hull
        self.scattered_wave_vector = scattered_wave_vector
        self.incident_wave_vector  = incident_wave_vector
        self.wavelength            = wavelength
        self.incident_polarization_vector = incident_polarization_vector
        self.rotation_axis         = rotation_axis
        self.time                  = time
        self.phase                 = phase
        self.hkl_indx              = hkl_indx

    @property
    def hkl(self):
        """hkl (:obj:`numpy array`): Miller indices [h,k,l] ``shape=(3,)``."""
        return self.phase.miller_indices[ self.hkl_indx ]

    @property
    def real_structure_factor(self):
        """hkl (:obj:`numpy array`): Real part of unit cell structure factor"""
        if self.phase.structure_factors is not None:
            return self.phase.structure_factors[ self.hkl_indx, 0 ]
        else:
            return self.phase.structure_factors

    @property
    def imaginary_structure_factor(self):
        """hkl (:obj:`numpy array`): Imaginary part of unit cell structure factor"""
        if self.phase.structure_factors is not None:
            return self.phase.structure_factors[ self.hkl_indx, 1 ]
        else:
            return self.phase.structure_factors

    @property
    def lorentz_factor( self ):
        """Compute the Lorentz intensity factor for a scatterer.
        """
        k           = self.incident_wave_vector 
        kp          = self.scattered_wave_vector
        theta       = np.arccos( k.dot( kp ) / (np.linalg.norm(k)**2) ) / 2.
        korthogonal = kp - k * kp.dot( k ) / np.linalg.norm(k**2)
        eta         = np.arccos( self.rotation_axis.dot( korthogonal ) / np.linalg.norm( korthogonal ) )
        if eta<1e-8 or theta<1e-8:
            return np.inf
        else:
            return 1. / ( np.sin(2 * theta) * np.abs( np.sin(eta) ) )

    @property
    def polarization_factor( self ):
        """Compute the Polarization intensity factor for a scatterer.
        """
        khatp = self.scattered_wave_vector  / np.linalg.norm( self.scattered_wave_vector  )
        return 1 - np.dot( self.incident_polarization_vector, khatp )**2

    @property
    def centroid(self):
        """centroid (:obj:`numpy array`): centroid of the scattering region. ``shape=(3,)``
        """
        return np.mean( self.convex_hull.points[self.convex_hull.vertices], axis=0 )

    @property
    def volume(self):
        """volume (:obj:`float`): volume of the scattering region volume 
        """
        return self.convex_hull.volume



import os
import unittest
import numpy as np
from xrd_simulator.phase import Phase
import pkg_resources

class TestPhase(unittest.TestCase):

    def test_init(self):
        unit_cell = [5.028, 5.028, 5.519, 90., 90., 120.]
        sgname = "P3221"
        ph  = Phase(unit_cell, sgname)
        self.assertTrue( ph.structure_factors is None )

    def test_setup_diffracting_planes(self):
            
        unit_cell = [5.028, 5.028, 5.519, 90., 90., 120.]
        sgname = "P3221"
        ph  = Phase(unit_cell, sgname)

        wavelength = 1.0
        min_bragg_angle = 1  * np.pi/180
        max_bragg_angle = 25 * np.pi/180

        ph.setup_diffracting_planes( wavelength, min_bragg_angle, max_bragg_angle )

        self.assertEqual(  ph.miller_indices.shape[1], 3 )
        self.assertTrue(   ph.miller_indices.shape[0] > 10 )

        unit_cell = [3.64570000, 3.64570000, 3.64570000, 90.0, 90.0, 90.0]
        sgname = 'Fm-3m' # Iron
        ph  = Phase(unit_cell, sgname)
        ph.setup_diffracting_planes( wavelength, min_bragg_angle, max_bragg_angle )
        for i in range(ph.miller_indices.shape[0]):
            h,k,l = ph.miller_indices[i,:]
            # Only all even or all odd gives diffraction for a cubic crystal
            self.assertTrue( (h%2==0 and k%2==0 and l%2==0) or (h%2==1 and k%2==1 and l%2==1) )

    def test_set_structure_factors(self):
        data = os.path.join( os.path.join(os.path.dirname(__file__), 'data' ), 'Fe_mp-150_conventional_standard.cif' )
        #data = pkg_resources.resource_filename(__name__,  "data/Fe_mp-150_conventional_standard.cif")
        print('data', data, __file__ )
        unit_cell = [3.64570000, 3.64570000, 3.64570000, 90.0, 90.0, 90.0]
        sgname = 'Fm-3m' # Iron
        ph   = Phase(unit_cell, sgname, path_to_cif_file=data)
        wavelength = 1.0
        min_bragg_angle = 1  * np.pi/180
        max_bragg_angle = 25 * np.pi/180
        ph.setup_diffracting_planes( wavelength, min_bragg_angle, max_bragg_angle )

        for i in range(ph.structure_factors.shape[0]):
            self.assertGreaterEqual(ph.structure_factors[i,0], 0)


if __name__ == '__main__':
    unittest.main()
